/* $EPIC:$ */
/*
 * dcc.c: Things dealing client to client connections. 
 *
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993-2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include "irc.h"
#include "crypt.h"
#include "ctcp.h"
#include "dcc.h"
#include "hook.h"
#include "ircaux.h"
#include "lastlog.h"
#include "output.h"
#include "parse.h"
#include "server.h"
#include "status.h"
#include "vars.h"
#include "window.h"
#include "term.h"
#include "notice.h"
#include "newio2.h"
#include "network.h"
#include <sys/stat.h>

#define DCC_BLOCK_SIZE	BIG_BUFFER_SIZE

/* These are the settings for ``flags'' */
#define DCC_CHAT	((unsigned) 0x0001)
#define DCC_FILEOFFER	((unsigned) 0x0002)
#define DCC_FILEREAD	((unsigned) 0x0003)
#define	DCC_RAW		((unsigned) 0x0004)
#define	DCC_RAW_LISTEN	((unsigned) 0x0005)
#define DCC_TYPES	((unsigned) 0x000f)
#define DCC_MY_OFFER	((unsigned) 0x0010)
#define DCC_ACTIVE	((unsigned) 0x0020)
#define DCC_THEIR_OFFER	((unsigned) 0x0040)
#define DCC_DELETE	((unsigned) 0x0080)
#define DCC_TWOCLIENTS	((unsigned) 0x0100)
#define DCC_REJECTED	((unsigned) 0x0200)
#define DCC_REVERSE	((unsigned) 0x0400)
#define DCC_STATES	((unsigned) 0xfff0)

#define DCC_TYPE(x) ((x) & DCC_TYPES)

typedef	struct	DCC_struct
{
		unsigned	flags;
		int		peer;
		int		file;
		unsigned long	filesize;
		char *		description;
		char *		filename;
		char *		user;
		char *		othername;
		char *		encrypt;
	struct	DCC_struct *	next;
		ISA		remote;
		ISA		local;
#define remote_addr remote.sin_addr
#define remote_port remote.sin_port
#define local_addr local.sin_addr
#define local_port local.sin_port
		off_t		bytes_read;
		off_t		bytes_sent;
		Timeval		lasttime;
		Timeval		starttime;
		char *		cksum;
		unsigned long	packets_total;
		unsigned long	packets_transfer;
		unsigned long	packets_ack;
		long		packets_outstanding;
		int		connection_in_progress;
}	DCC_list;

static	off_t		filesize = 0;
static	DCC_list *	ClientList = NULL;
static	char		DCC_current_transfer_buffer[256];
	time_t		dcc_timeout = 600;		/* Backed by a /set */

static	void 		dcc_add_deadclient 	(DCC_list *);
static	void		dcc_chat 		(char *);
static	void 		dcc_close 		(char *);
static	void		dcc_closeall		(char *);
static	int		dcc_erase 		(DCC_list *);
static	void		dcc_filesend 		(char *);
static	void		dcc_getfile 		(char *);
static	int		dcc_open 		(DCC_list *);
static	void 		dcc_really_erase 	(void);
static	void		dcc_rename 		(char *);
static	void		dcc_reverse		(char *);
static	DCC_list *	dcc_searchlist 		(const char *, const char *, unsigned, int, const char *, int);
static	void		dcc_send_raw 		(char *);
static	int		dcc_send_packets 	(DCC_list *Client);
static	void 		output_reject_ctcp 	(char *, char *);
static	void		DCC_close_filesend 	(DCC_list *, char *);
static	void		update_transfer_buffer 	(char *format, ...);
static 	void		dcc_send_booster_ctcp 	(DCC_list *dcc);
static	char *		dcc_urlencode		(const char *);
static	char *		dcc_urldecode		(const char *);

#ifdef MIRC_BROKEN_DCC_RESUME
static	void		dcc_getfile_resume 	    (char *);
static 	void 		dcc_getfile_resume_demanded (char *user, char *filename, char *port, char *offset);
static	void		dcc_getfile_resume_start    (char *nick, char *filename, char *port, char *offset);
#endif

static	int		dcc_chat_reader	(int, int, char *);
static	int		dcc_chat_error	(int);
static	int		dcc_chat_accept	(int, int, char *);
static	Discipline	dcc_chat_discipline = {
		dcc_chat_reader,
		NULL,
		dcc_chat_error,
		dcc_chat_error
	};

static	int		dcc_raw_error  (int);
static	int		dcc_raw_reader (int, int, char *);
static	int		dcc_raw_accept (int, int, char *);
static	Discipline	dcc_raw_discipline = {
		dcc_raw_reader,
		NULL,
		dcc_raw_error,
		dcc_raw_error
	};

static	int		dcc_send_accept (int, int, char *);
static	int		dcc_send_reader (int, int, char *);
static	int		dcc_send_error  (int);
static	Discipline	dcc_send_discipline = {
		dcc_send_reader,
		NULL,
		dcc_send_error,
		dcc_send_error
	};

static	int		dcc_get_reader (int, int, char *);
static	int		dcc_get_error  (int);
static	Discipline	dcc_get_discipline = {
		dcc_get_reader,
		NULL,
		dcc_get_error,
		dcc_get_error
	};


/*
 * These are the possible <type> arguments to /DCC
 */
typedef void (*dcc_function) (char *);
struct
{
	char	*	name;
	dcc_function 	function;
}	dcc_commands[] =
{
	{ "CHAT",	dcc_chat 		},	/* DCC_CHAT */
	{ "SEND",	dcc_filesend 		},	/* DCC_FILEOFFER */
	{ "GET",	dcc_getfile 		},	/* DCC_FILEREAD */
	{ "RAW",	dcc_send_raw 		},	/* DCC_RAW */

	{ "CLOSE",	dcc_close 		},
	{ "CLOSEALL",	dcc_closeall		},
	{ "LIST",	dcc_list 		},
	{ "RENAME",	dcc_rename 		},
	{ "REVERSE",	dcc_reverse		},
#ifdef MIRC_BROKEN_DCC_RESUME
	{ "RESUME",	dcc_getfile_resume 	},
#endif
	{ NULL,		(dcc_function) NULL 	}
};

/*
 * These are the possible types of DCCs that can be open.
 */
char	*dcc_types[] =
{
	"<none>",
	"CHAT",		/* DCC_CHAT */
	"SEND",		/* DCC_FILEOFFER */
	"GET",		/* DCC_FILEREAD */
	"RAW",		/* DCC_RAW */
	"RAW_LISTEN",
	NULL
};

struct	deadlist
{
	DCC_list *	it;
	struct deadlist *next;
}	*deadlist = NULL;


/*
 * remove_from_dcc_list: What do you think it does?
 */
static void	dcc_remove_from_list (DCC_list *erased)
{
	DCC_list *prev = NULL;

	if (erased != ClientList)
	{
		for (prev = ClientList; prev; prev = prev->next)
			if (prev->next == erased)
				break;
	}

	if (erased == ClientList)
		ClientList = erased->next;
	else if (prev)
		prev->next = erased->next;
}

/*
 * When a DCC is considered to be "invalid", the DCC_DELETE bit is set.
 * The structure is not immediately removed, however -- it sticks around
 * until the next looping synchornization point where dcc_check() collects
 * the "deleted" DCC structures into the 'deadlist'.  After all of the
 * DCCs have been parsed, it then goes through and cleans the list.  This
 * ensures that no DCCs are removed from under the feet of some unsuspecting
 * function -- they are only deleted synchronously.
 *
 * The downside is that it requires an M*N algorithm to remove the DCCs.
 */
static void 	dcc_add_deadclient (DCC_list *client)
{
	struct deadlist *new_dl;

	for (new_dl = deadlist; new_dl; new_dl = new_dl->next)
		if (new_dl->it == client)
			return;

	new_dl = new_malloc(sizeof(struct deadlist));
	new_dl->next = deadlist;
	new_dl->it = client;
	deadlist = new_dl;
}

static void 	dcc_really_erase (void)
{
	struct deadlist *dies;

	while ((dies = deadlist))
	{
		deadlist = deadlist->next;
		dcc_erase(dies->it);
		new_free((char **)&dies);
	}
	update_transfer_buffer(NULL);	/* Whatever */
	update_all_status();
}

/*
 * Note that 'erased' does not neccesarily have to be on ClientList.
 * In fact, it may very well NOT be on ClientList.  The handling at the
 * beginning of the function is only to sanity check that it isnt on 
 * ClientList when we blow it away.
 */
static 	int		dcc_erase (DCC_list *erased)
{
	int	retval = 0;

	/*
	 * Automatically grok for REJECTs
	 */
	if (!(erased->flags & DCC_REJECTED))
	{
	    if (erased->flags & (DCC_MY_OFFER|DCC_THEIR_OFFER))
	     if (get_int_var(DCC_AUTO_SEND_REJECTS_VAR))
              do
	      {
		unsigned my_type = DCC_TYPE(erased->flags);
		char *		dummy_ptr = NULL;
		char *		nopath;
		char *		dummy_nick = NULL;
		static time_t	last_reject = 0;
		time_t		now;

		time(&now);
		if (now - last_reject < 2)
			break;		/* Throttle flood attempts */
		last_reject = now;

		if (erased->description && 
				(nopath = strrchr(erased->description, '/')))
			nopath++;
		else
			nopath = erased->description;

		if (*(erased->user) == '=')
			dummy_nick = erased->user + 1;
		else
			dummy_nick = erased->user;

		dummy_ptr = m_sprintf("%s %s %s", 
			dummy_nick,
			(my_type == DCC_FILEOFFER ? "GET" :
			 (my_type == DCC_FILEREAD  ? "SEND" : 
			   dcc_types[my_type])),
			nopath ? nopath : "<any>");

		erased->flags |= DCC_REJECTED;
		retval = 1;

		/*
		 * And output it to the user
		 */
		if (!dead)
			isonbase(dummy_ptr, output_reject_ctcp);
		else
			output_reject_ctcp(dummy_ptr, erased->user);
	      }
	      while (0);
	}

	dcc_remove_from_list(erased);

	/*
	 * In any event, blow it away.
	 */
	erased->peer = new_close(erased->peer);
	new_free(&erased->description);	/* Magic check failure here */
	new_free(&erased->filename);
	new_free(&erased->user);
	new_free(&erased->othername);
	new_free(&erased->encrypt);
	new_free(&erased->cksum);
	new_free((char **)&erased);
	return retval;
}

/*
 * close_all_dcc:  We call this when we create a new process so that
 * we don't leave any fd's lying around, that won't close when we
 * want them to..
 */
void 	close_all_dcc (void)
{
	DCC_list *Client;
	int	dawdle = 0;

	while ((Client = ClientList))
		if (dcc_erase(Client) && dead)
			dawdle = 1;

	if (dawdle)
	{
		say("Waiting for DCC REJECTs to be sent");
		sleep(1);
	}
}




/*
 * These functions handle important DCC jobs.
 */

/*
 * dcc_searchlist searches through the dcc_list and finds the client
 * with the flag described in type set.  This function should never
 * return a delete'd entry.
 */
static	DCC_list *dcc_searchlist (
	const char *	description,	/* 
					 * DCC Type specific information,
					 * Usually a full pathname for 
					 * SEND/GET, "listen" or "connect"
					 * for RAW, or NULL for others.
					 */
	const char *	user, 		/* Nick of the remote peer */
	unsigned 	type, 		/* Flags, including type of conn */
	int 		create, 	/* Create if it doesnt exist? */
	const char *	othername, 	/* Alias filename for SEND/GET */
	int 		active)		/* Only get active/non-active? */
{
	DCC_list 	*client, 
			*new_client;
	const char 	*last = NULL;

	if (x_debug & DEBUG_DCC_SEARCH)
		yell("entering dcc_sl.  desc (%s) user (%s) type (%d) "
		     "create (%d) other (%s) active (%d)", 
			description, user, type, create, othername, active);

	/*
	 * Walk all of the DCC connections that we know about.
	 */
	for (client = ClientList; client ; client = client->next)
	{
		/*
		 * Tell the user if they care
		 */
		if (x_debug & DEBUG_DCC_SEARCH)
		{
			yell("checking against  name (%s) user (%s) type (%d) "
			     "flag (%d) other (%s) active (%x)", 
				client->description, 
				client->user, 
				DCC_TYPE(client->flags),
				client->flags, 
				client->othername, 
				client->flags & DCC_ACTIVE);
		}


		/*
		 * Ok. first of all, it has to be the right type.
		 */
		if (type != -1 && (DCC_TYPE(client->flags) != type))
			continue;

		/*
		 * Its OK if the user matches the client's user
		 */
		if (user && my_stricmp(user, client->user))
			continue;


		/*
		 * If "name" is NULL, then that acts as a wildcard.
		 * If "description" is NULL, then that also acts as a wildcard.
		 * If "name" is not the same as "description", then it could
		 * 	be that "description" is a filename.  Check to see if
		 *	"name" is the last component in "description" and
		 *	accept that.
		 * Otherwise, "othername" must exist and be the same.
		 * In all other cases, reject this entry.
		 */
		if (description && client->description && 
			my_stricmp(description, client->description))
		{
			/*
			 * OK.  well, 'name' isnt 'description', try looking
			 * for a last segment.  If there isnt one, choke.
			 */
			if ((last = strrchr(client->description, '/')) == NULL)
				continue;

			/*
			 * If 'name' isnt 'last' then we'll have to look at
			 * 'othername' to see if that matches.
			 */
			if (my_stricmp(description, last + 1))
			{
				if (!othername || !client->othername)
					continue;

				if (my_stricmp(othername, client->othername))
					continue;
			}
		}

		/* Never return deleted entries */
		if (client->flags & DCC_DELETE)
			continue;

		/*
		 * Active == 0  -> Only PENDING unopened connections, please
		 * No deleted entries, either.
		 */
		if (active == 0)
		{
			if (client->flags & DCC_ACTIVE)
				continue;
		}
		/*
		 * Active == 1 -> Only ACTIVE and OPEN connections, please
		 */
		else if (active == 1)
		{
			if ((client->flags & DCC_ACTIVE) == 0)
				continue;
		}
		/*
		 * Active == -1 -> Only NON DELETED entries, please.
		 */
		else if (active == -1)
			;

		if (x_debug & DEBUG_DCC_SEARCH)
			yell("We have a winner!");

		/* Looks like we have a winner! */
		return client;
	}

	/*
	 * If flag == 1, then the user wants us to create a new DCC
	 * client if neccesary.  Otherwise, they were just checking to see
	 * if one existed and dont want to create it.
	 */
	if (!create)
		return NULL;

	new_client 			= new_malloc(sizeof(DCC_list));
	new_client->flags 		= type;
	new_client->peer 		= -1;
	new_client->file 		= -1;
	new_client->filesize 		= filesize;
	new_client->filename 		= NULL;
	new_client->packets_total 	= filesize ? 
					  (filesize / DCC_BLOCK_SIZE + 1) : 0;
	new_client->packets_transfer 	= 0;
	new_client->packets_outstanding = 0;
	new_client->packets_ack 	= 0;
	new_client->next 		= ClientList;
	new_client->user 		= m_strdup(user);
	new_client->description 	= m_strdup(description);
	new_client->othername 		= m_strdup(othername);
	new_client->bytes_read 		= 0;
	new_client->bytes_sent 		= 0;
	new_client->starttime.tv_sec 	= 0;
	new_client->starttime.tv_usec 	= 0;
	new_client->local_port 		= 0;
	new_client->cksum 		= NULL;
	new_client->encrypt 		= NULL;
	new_client->connection_in_progress = 0;
	get_time(&new_client->lasttime);

	ClientList = new_client;
	return new_client;
}

/*
 * Added by Chaos: Is used in edit.c for checking redirect.
 */
int	dcc_chat_active (const char *user)
{
	return (dcc_searchlist("chat", user, DCC_CHAT, 0, NULL, 1)) ? 1 : 0;
}

static	int	dcc_nonblock_callback (int fd, int len, char *buffer)
{
	DCC_list *client;

	for (client = ClientList; client ; client = client->next)
	{
		if (client->peer == fd)
		{
			client->connection_in_progress = 0;
			return 0;
		}
	}
	panic("dcc_nonblock_callback called spuriously.");
}

static	int	dcc_open_outbound (DCC_list *dcc)
{
	char *		user;
	char *		type;
	ISA		remaddr;
	int		rl = sizeof(remaddr);
	int		old_server;
	int		jvs_blah = 1;

	/*
	 * Initialize our idea of what is going on.
	 */
	user = dcc->user;
	old_server = from_server;
	if (from_server == -1)
		from_server = get_window_server(0);
	type = dcc_types[DCC_TYPE(dcc->flags)];


	/*
	 * Connect to them.
	 */
	/* XXXX -- hack -- XXXX */
	dcc->connection_in_progress = 1;
	dcc->remote_port = ntohs(dcc->remote_port);
	if ((dcc->peer = connect_by_number(inet_ntoa(dcc->remote_addr), 
					    &dcc->remote_port, 
					    SERVICE_CLIENT, 
					    PROTOCOL_TCP, 
					    dcc_nonblock_callback)) < 0)
	{
		dcc->flags |= DCC_DELETE;

		set_display_target(NULL, LOG_DCC);
		say("Unable to create connection: [%d] %s", 
			errno, 
			errno ? strerror(errno) : "Unknown Host");
		reset_display_target();

		dcc->peer = -1;
		from_server = old_server;
		return 0;
	}
	dcc->remote_port = htons(dcc->remote_port);

	/* Wait for the connect */
	while (dcc->connection_in_progress)
		io("Waiting on dcc connect");

	/*
	 * Who is on the other end?
	 * Whine if it isn't who we thought we connected to.
	 * And in any event use the real connection address.
	 * XXX Unfortunately, the 'localhost' address can be mapped
	 * to more than one internet address (0.0.0.0 as well as 127.0.0.1)
	 */
	if (getpeername(dcc->peer, (SA *) &remaddr, &rl))
	{
		dcc->flags |= DCC_DELETE;

		say("Unable to create connection: [%d] %s",
			errno,
			errno ? strerror(errno) : "Unknown Host");
		reset_display_target();
		dcc->peer = new_close(dcc->peer);
		from_server = old_server;
		return 0;
	}

	if (SOCKADDR(remaddr) != ADDRESS(dcc->remote_addr))
	{
	    if (SOCKADDR(remaddr) && ADDRESS(dcc->remote_addr))
	    {
		say("The address we connected to [%s (%d)] is different "
		    " from the one we tried to connect to [%s (%d)]!  ",
			inet_ntoa(remaddr.sin_addr), SOCKADDR(remaddr),
			inet_ntoa(dcc->remote_addr), ADDRESS(dcc->remote_addr));
		say("Be Careful!");
	    }
	}

	/*
	 * Set up the connection to be useful
	 */
	if (DCC_TYPE(dcc->flags) == DCC_CHAT)
		line_buffering(dcc->peer, &dcc_chat_discipline);
	else if (DCC_TYPE(dcc->flags) == DCC_FILEREAD)
		no_buffering_readonly(dcc->peer, &dcc_get_discipline);
	else if (DCC_TYPE(dcc->flags) == DCC_RAW)
		line_buffering(dcc->peer, &dcc_raw_discipline);
	else if (DCC_TYPE(dcc->flags) == DCC_FILEOFFER)
		no_buffering_readonly(dcc->peer, &dcc_send_discipline);

	dcc->flags &= ~DCC_THEIR_OFFER;
	dcc->flags &= ~DCC_MY_OFFER;
	dcc->flags |= DCC_ACTIVE;


	dcc->remote_addr = remaddr.sin_addr;

	/*
	 * If this was a two-peer connection, then tell the user
	 * that the connection was successfull.
	 */
	if (DCC_TYPE(dcc->flags) != DCC_RAW)
	{
		set_display_target(NULL, LOG_DCC);

		/* 
		 * It would be nice if SEND also showed the filename
		 * and size (since it's possible to have multiple
		 * SEND requests queued), so we check for a file size
		 * first. 
		 * Actually, that was wrong.  We just check the type.
		 * so that it doesnt choke on zero-length files.
		 */
		if (DCC_TYPE(dcc->flags) == DCC_FILEREAD ||
		   (DCC_TYPE(dcc->flags) == DCC_FILEOFFER))
		{
		    if ((jvs_blah = do_hook(DCC_CONNECT_LIST,
					"%s %s %s %d %s %ld", 
					user, type,
					inet_ntoa(dcc->remote_addr),
					ntohs(dcc->remote_port),
					dcc->description,
					dcc->filesize)))
		    /*
		     * Compatability with bitchx
		     */
			jvs_blah = do_hook(DCC_CONNECT_LIST,
					"%s GET %s %d %s %ld", 
					user, inet_ntoa(dcc->remote_addr),
					ntohs(dcc->remote_port),
					dcc->description,
					dcc->filesize);
		}
		else
		{
		    jvs_blah = do_hook(DCC_CONNECT_LIST,
					"%s %s %s %d", 
					user, type,
					inet_ntoa(dcc->remote_addr),
					ntohs(dcc->remote_port));
		}

		if (jvs_blah)
		{
		    say("DCC %s connection with %s[%s:%d] established",
				type, user, 
				inet_ntoa(dcc->remote_addr),
				ntohs(dcc->remote_port));
		}
		reset_display_target();
	}

	/*
	 * Record the time the connection was started, 
	 * Clean up and then return.
	 */
	get_time(&dcc->starttime);
	from_server = old_server;
	return 1;
}

static	int		dcc_open_inbound (DCC_list *dcc)
{
	char *		user;
	char *		type;
	int		old_server;
	unsigned short 	portnum1 = 0;
	unsigned short	portnum2 = 0;

	/*
	 * Initialize our idea of what is going on.
	 */
	user = dcc->user;
	old_server = from_server;
	if (from_server == -1)
		from_server = get_window_server(0);
	type = dcc_types[DCC_TYPE(dcc->flags)];


	if (get_int_var(RANDOM_LOCAL_PORTS_VAR))
		portnum1 = random_number(65535 - 1024) + 1024;

	/*
	 * Mark that we're waiting for the remote peer to answer,
	 * and then open up a listen()ing socket for them.  If our
	 * first choice of port fails, try another one.  If both
	 * fail, then we give up.  If the user insists on doing
	 * random ports, then we will fallback to asking the system
	 * for a port if our random port isnt available.
	 */
	dcc->flags |= DCC_MY_OFFER;
	if ((dcc->peer = connect_by_number(NULL, &portnum1, 
				SERVICE_SERVER, PROTOCOL_TCP, NULL)) < 0)
	{
	    if ((dcc->peer = connect_by_number(NULL, &portnum2,
				SERVICE_SERVER, PROTOCOL_TCP, NULL)) < 0)
	    {
		dcc->flags |= DCC_DELETE;

		set_display_target(NULL, LOG_DCC);
		say("Unable to create connection: %s", 
			errno ? strerror(errno) : "Unknown Host");
		reset_display_target();

		dcc->peer = -1;
		from_server = old_server;
		return 0;
	    }
	    else
		dcc->local_port = portnum2;
	}
	else
	    dcc->local_port = portnum1;

#ifdef MIRC_BROKEN_DCC_RESUME
	/*
	 * For stupid MIRC dcc resumes, we need to stash the
	 * local port number, because the remote client will send
	 * back that port number as its ID of what file it wants
	 * to resume (rather than the filename. ick.)
	 */
	malloc_strcpy(&dcc->othername, ltoa(dcc->local_port));
#endif

	if (DCC_TYPE(dcc->flags) == DCC_CHAT)
		accept_callback(dcc->peer, dcc_chat_accept);
	else if (DCC_TYPE(dcc->flags) == DCC_FILEOFFER)
		accept_callback(dcc->peer, dcc_send_accept);
	else if (DCC_TYPE(dcc->flags) == DCC_RAW_LISTEN)
		accept_callback(dcc->peer, dcc_raw_accept);

	/*
	 * If this is to be a 2-peer connection, then we need to
	 * send the remote peer a CTCP request.  I suppose we should
	 * do an ISON request first, but thats another project.
	 */
	if (dcc->flags & DCC_TWOCLIENTS)
		dcc_send_booster_ctcp(dcc);

	from_server = old_server;
	return 2;
}

/*
 * Whenever a DCC changes state from WAITING->ACTIVE, it calls this function
 * to initiate the internet connection for the transaction.
 */
static	int		dcc_open (DCC_list *dcc)
{
	/*
	 * If this is ME accepting someone ELSES offer...
	 */
	if (dcc->flags & DCC_THEIR_OFFER)
		return dcc_open_outbound(dcc);

	/*
	 * The user is asking us to set up a connection for a remote
	 * peer to connect to US.
	 */
	else
		return dcc_open_inbound(dcc);
}

/*
 * send_booster_ctcp: This is called by dcc_open and also by dcc_filesend
 * to send a CTCP handshake message to a remote peer.  The reason its a 
 * function is because its a rather large chunk of code, and it needs to be
 * done basically identically by both places.  Whatever.
 */
static void	dcc_send_booster_ctcp (DCC_list *dcc)
{
	char		*nopath;
	char		*type = dcc_types[dcc->flags & DCC_TYPES];
	IA		myip;
	char		*reverse = empty_string;

	if (get_int_var(DCC_USE_GATEWAY_ADDR_VAR))
		ADDRESS(myip) = ADDRESS(get_server_uh_addr(from_server));
	else
		ADDRESS(myip) = ADDRESS(get_server_local_addr(from_server));

	/*
	 * If this is to be a 2-peer connection, then we need to
	 * send the remote peer a CTCP request.  I suppose we should
	 * do an ISON request first, but thats another project.
	 */
	if (!(dcc->flags & DCC_TWOCLIENTS))
		return;

	get_time(&dcc->starttime);
	get_time(&dcc->lasttime);

	if ((dcc->flags & DCC_FILEOFFER) && 
		  (nopath = strrchr(dcc->description, '/')))
		nopath++;
	else
		nopath = dcc->description;

	/*
	 * If this is a DCC SEND...
	 */
	if (dcc->flags & DCC_FILEOFFER)
	{
		char *	url_name = dcc_urlencode(nopath);

		/*
		 * Dont bother with the checksum any more.
		 */
		send_ctcp(CTCP_PRIVMSG, dcc->user, CTCP_DCC,
			 "%s %s %lu %u %ld", 
			 type, url_name,
			 (u_long) ntohl(ADDRESS(myip)),
			 (u_short) dcc->local_port,
			 dcc->filesize);

		/*
		 * Tell the user we sent out the request
		 */
		set_display_target(NULL, LOG_DCC);
		if (do_hook(DCC_OFFER_LIST, "%s %s%s %s %ld", 
			dcc->user, reverse, type, url_name, dcc->filesize))
		    say("Sent DCC %s%s request (%s %ld) to %s", 
			reverse, type, nopath, dcc->filesize, dcc->user);
		reset_display_target();

		new_free(&url_name);
	}

	/*
	 * Otherwise its a DCC CHAT request
	 */
	else
	{
		/*
		 * Send out the handshake
		 */
		send_ctcp(CTCP_PRIVMSG, dcc->user, CTCP_DCC,
			 "%s%s %s %lu %u", 
			 reverse, type, nopath,
			 (u_long) ntohl(ADDRESS(myip)),
			 (u_short) dcc->local_port);

		/*
		 * And tell the user
		 */
		set_display_target(NULL, LOG_DCC);
		if (do_hook(DCC_OFFER_LIST, "%s %s%s", 
			dcc->user, reverse, type))
		    say("Sent DCC %s%s request to %s", 
			reverse, type, dcc->user);
		reset_display_target();
	}
}


/*
 * This allows you to send (via /msg or /query) a message to a remote 
 * dcc target.  The two types of targets you may send to are a DCC CHAT
 * connection (you specify a nickname) or a DCC RAW connection (you specify
 * the file descriptor).  Of course, since DCC RAW connections use digits
 * and DCC CHATs dont, sending a message to =<digits> then is presumed to
 * be a message to a DCC RAW.
 *
 * If ``noisy'' is 1, then we tell the user that we send the message, and
 * allow them to hook it.  If ``noisy'' is 0, then we're doing this silently
 * (as in /redirect, or /ctcp) and we dont want to tell the user about it.
 *
 * XXXX This needs to be broken up into CHAT and RAW cases.
 */
static void	dcc_message_transmit (
	int 		type, 		/* What type of DCC to send over */
	char 		*user, 		/* Who to send it to */
	char		*desc,		/* THe 'desc' field in DCC_List */
	char 		*text, 		/* What to send them */
const	char 		*text_display, 	/* What to tell the user we sent */
	int 		noisy, 		/* Do we tell the user? */
	const char 	*cmd)		/* 
					 * For /CTCP's, is this a PRIVMSG
					 * or a NOTICE?
					 */
{
	DCC_list	*dcc;
	char		tmp[DCC_BLOCK_SIZE+1];
	char		thing = '\0';
	int		list = 0;
	int		old_from_server = from_server;

	tmp[0] = 0;

	switch (type)
	{
		case DCC_CHAT:
		{
			thing = '=';
			list = SEND_DCC_CHAT_LIST;
			break;
		}
		case DCC_RAW:
		{
			noisy = 0;
			break;
		}
	}

	if (!(dcc = dcc_searchlist(desc, user, type, 0, NULL, -1)))
	{
		say("No active DCC %s:%s connection for %s",
			dcc_types[type], desc, user);
		return;
	}


	/*
	 * Check for CTCPs... whee.
	 * Dont tag outbound ACTIONs, which break mirc.  Blah.
	 */
	if (*text == CTCP_DELIM_CHAR && strncmp(text + 1, "ACTION", 6))
	{
		if (!strcmp(cmd, "PRIVMSG"))
			strmcpy(tmp, "CTCP_MESSAGE ", DCC_BLOCK_SIZE);
		else
			strmcpy(tmp, "CTCP_REPLY ", DCC_BLOCK_SIZE);
	}

	strmcat(tmp, text, DCC_BLOCK_SIZE);
	strmcat(tmp, "\n", DCC_BLOCK_SIZE); 

	set_display_target(dcc->user, LOG_DCC);

	if (x_debug & DEBUG_OUTBOUND) 
		yell("-> [%s] [%s]", desc, tmp);

	if (write(dcc->peer, tmp, strlen(tmp)) == -1)
	{
		dcc->flags |= DCC_DELETE;

		say("Outbound write() failed: %s", 
			errno ? strerror(errno) : 
			"What the heck is wrong with your disk?");

		reset_display_target();
		from_server = old_from_server;
		return;
	}

	dcc->bytes_sent += strlen(tmp);

	if (noisy)
	{
		if (do_hook(list, "%s %s", dcc->user, text_display))
			put_it("=> %c%s%c %s", 
				thing, dcc->user, thing, text_display);
	}

	reset_display_target();
	return;
}

/*
 * This is used to send a message to a remote DCC peer.  This is called
 * by send_text.
 */
void	dcc_chat_transmit (char *user, char *text, const char *orig, const char *type, int noisy)
{
	int	fd;

	/*
	 * This converts a message being sent to a number into whatever
	 * its local port is (which is what we think the nickname is).
	 * Its just a 15 minute hack. dont read too much into it.
	 */
	if ((fd = atol(user)))
	{
		DCC_list *	dcc;

		for (dcc = ClientList; dcc; dcc = dcc->next)
			if (dcc->peer == fd)
				break;

		if (!dcc)
		{
			put_it("Descriptor %d is not an open DCC RAW", fd);
			return;
		}

		dcc_message_transmit(DCC_RAW, dcc->user, dcc->description, 
					text, orig, noisy, type);
	}
	else
		dcc_message_transmit(DCC_CHAT, user, "chat",
					text, orig, noisy, type);
}




/*
 *
 * All these functions are user-generated -- that is, they are all called
 * when the user does /DCC <command> or one of the DCC-oriented built in
 * functions.
 *
 */


/*
 * The /DCC command.  Delegates the work to other functions.
 */
void	process_dcc(char *args)
{
	char	*command;
	int	i;

	if (!(command = next_arg(args, &args)))
		return;

	for (i = 0; dcc_commands[i].name != NULL; i++)
	{
		if (!my_stricmp(dcc_commands[i].name, command))
		{
			set_display_target(NULL, LOG_DCC);
			dcc_commands[i].function(args);
			reset_display_target();
			return;
		}
	}
	say("Unknown DCC command: %s", command);
}


/*
 * Usage: /DCC CHAT <nick> [-e passkey]
 */
static void	dcc_chat (char *args)
{
	char		*user;
	DCC_list	*dcc;
	int		portnum = 0;

	if ((user = next_arg(args, &args)) == NULL)
	{
		say("You must supply a nickname for DCC CHAT");
		return;
	}

	/*
	 * Check to see if there is a flag
	 */
	if (*args == '-')
	{
		if (args[1] == 'p')
		{
			if (args && *args)
			    portnum = my_atol(next_arg(args, &args));
		}
	}

	dcc = dcc_searchlist("chat", user, DCC_CHAT, 1, NULL, -1);
	if ((dcc->flags & DCC_ACTIVE) || (dcc->flags & DCC_MY_OFFER))
	{
		say("A previous DCC CHAT to %s exists", user);
		return;
	}

	dcc->flags |= DCC_TWOCLIENTS;
	dcc->local_port = portnum;
	dcc_open(dcc);
}

/*
 * Usage: /DCC CLOSE <type> <nick>
 */
static	void 	dcc_close (char *args)
{
	DCC_list	*dcc;
	char		*type;
	char		*user;
	char		*file;
	int		any_type = 0;
	int		any_user = 0;
	int		i;
	int		count = 0;

	type = next_arg(args, &args);
	user = next_arg(args, &args);
	file = next_arg(args, &args);

	if (type && (!my_stricmp(type, "-all") || !my_stricmp(type, "*")))
		any_type = 1;
	if (user && (!my_stricmp(user, "-all") || !my_stricmp(type, "*")))
		any_user = 1;

	if (!type || (!user && !any_type))
	{
		say("You must specify a type and a nick for DCC CLOSE");
		return;
	}

	if (!any_type)
	{
		for (i = 0; dcc_types[i]; i++)
			if (!my_stricmp(type, dcc_types[i]))
				break;

		if (!dcc_types[i])
		{
			say("Unknown DCC type: %s", type);
			return;
		}
	}
	else
		i = -1;

	if (any_user)
		user = NULL;

	while ((dcc = dcc_searchlist(file, user, i, 0, file, -1)))
	{
		unsigned my_type = dcc->flags & DCC_TYPES;

		count++;
		dcc->flags |= DCC_DELETE;

                if (do_hook(DCC_LOST_LIST,"%s %s %s USER ABORTED CONNECTION",
			dcc->user, 
			dcc_types[my_type],
                        dcc->description ? dcc->description : "<any>"))
		    say("DCC %s:%s to %s closed", 
			dcc_types[my_type],
			file ? file : "<any>", 
			dcc->user);
	}

	if (!count)
		say("No DCC %s:%s to %s found", 
			(i == -1 ? "<any>" : type), 
			(file ? file : "<any>"), 
			user);
}

/*
 * Usage: /DCC CLOSEALL
 * It leaves your DCC list very empty
 */
static void	dcc_closeall (char *args)
{
	char	*my_local_args = "-all -all";
	char	*breakage = LOCAL_COPY(my_local_args);

	dcc_close(breakage);	/* XXX Bleh */
}

/*
 * Usage: /DCC GET <nick> [file|*]
 * The '*' file gets all offered files.
 */
static	void	dcc_getfile (char *args)
{
	char		*user;
	char		*filename = NULL;
	DCC_list	*dcc;
	char		*fullname = NULL;
	char		*realname = NULL;
	char		pathname[MAXPATHLEN * 2 + 1];
	int		get_all = 0;
	int		count = 0;

	if (!(user = next_arg(args, &args)))
	{
		say("You must supply a nickname for DCC GET");
		return;
	}

	filename = next_arg(args, &args);
	if (filename && *filename && !strcmp(filename, "*"))
		get_all = 1, filename = NULL;

	while ((dcc = dcc_searchlist(filename, user, DCC_FILEREAD, 0, NULL, 0)))
	{
		count++;

		if (!get_all && dcc->flags & DCC_ACTIVE)
		{
			say("A previous DCC GET:%s to %s exists", 
				filename ? filename : "<any>", user);
			return;
		}
		if (!get_all && !(dcc->flags & DCC_THEIR_OFFER))
		{
			dcc->flags |= DCC_DELETE;
			say("I'm a little teapot");
			return;
		}

		dcc->flags |= DCC_TWOCLIENTS;
		if (!dcc_open(dcc))
		{
			if (get_all)
				continue;
			else
				return;
		}

		*pathname = 0;
		if (get_string_var(DCC_STORE_PATH_VAR))
		{
			strmcpy(pathname, get_string_var(DCC_STORE_PATH_VAR), 
						MAXPATHLEN * 2);
			strlcat(pathname, "/", MAXPATHLEN * 2);
		}

		realname = dcc_urldecode(dcc->description);
		strlcat(pathname, realname, MAXPATHLEN * 2);
		new_free(&realname);

		if (!(fullname = expand_twiddle(pathname)))
			malloc_strcpy(&fullname, pathname);

		dcc->filename = fullname;
		if ((dcc->file = open(fullname, 
				O_WRONLY|O_TRUNC|O_CREAT, 0644)) == -1)
		{
			dcc->flags |= DCC_DELETE;
			say("Unable to open %s: %s", 
				fullname, errno ? 
					strerror(errno) : 
					"<No Error>");
		}

		if (!get_all)
			break;
	}

	if (!count)
	{
		if (filename)
			say("No file (%s) offered in SEND mode by %s", 
					filename, user);
		else
			say("No file offered in SEND mode by %s", user);
		return;
	}
}



/*
 * Usage: /DCC LIST
 */
void	dcc_list (char *args)
{
	DCC_list	*Client;
static	char		*format =
		"%-7.7s%-3.3s %-9.9s %-9.9s %-20.20s %-8.8s %-8.8s %s";

	if (do_hook(DCC_LIST_LIST, "Start * * * * * * *"))
	{
		put_it(format, "Type", " ", "Nick", "Status", "Start time", 
				"Size", "Complete", "Args");
	}

	for (Client = ClientList ; Client != NULL ; Client = Client->next)
	{
		unsigned	flags = Client->flags;

		if (do_hook(DCC_LIST_LIST, "%s %s %s %s %ld %ld %ld %s",
				dcc_types[flags & DCC_TYPES],
				Client->encrypt ? one : zero,
				Client->user,
					flags & DCC_DELETE       ? "Closed"  :
					flags & DCC_ACTIVE       ? "Active"  :
					flags & DCC_MY_OFFER     ? "Waiting" :
					flags & DCC_THEIR_OFFER  ? "Offered" :
							           "Unknown",
				Client->starttime.tv_sec,
			  (long)Client->filesize,
				Client->bytes_sent ? (u_long)Client->bytes_sent
						   : (u_long)Client->bytes_read,
				Client->description))
		{
			char 	*filename = strrchr(Client->description, '/');
			char	completed[9];
			char	size[9];
			char	buf[23];
			char	*time_f;

			/*
			 * Figure out something sane for the filename
			 */
			if (!filename || get_int_var(DCC_LONG_PATHNAMES_VAR))
				filename = Client->description;
			else if (filename && *filename)
				filename++;

			if (!filename)
				filename = "*";

			/*
			 * Figure out something sane for the "completed"
			 */
			if (Client->filesize)
			{
				sprintf(completed, "%ld%%", 
					(u_long)(Client->bytes_sent
					         ? Client->bytes_sent
					         : Client->bytes_read) 
						* 100 / Client->filesize);
				strmcpy(size, ltoa(Client->filesize), 8);
			}
			else
			{
				sprintf(completed, "%ldK", 
					(u_long)(Client->bytes_sent 
					         ? Client->bytes_sent
					         : Client->bytes_read) / 1024);
				*size = 0;
			}


			/*
			 * Figure out something sane for starting time
			 */
			if (Client->starttime.tv_sec)
			{
				time_t 	blech = Client->starttime.tv_sec;
				struct tm *btime = localtime(&blech);

				strftime(buf, 22, "%T %b %d %Y", btime);
				time_f = buf;
			}
			else
				time_f = empty_string;

			/*
			 * And do the dirty work.
			 */
			put_it(format,
				dcc_types[flags&DCC_TYPES],
				Client->encrypt ? "[E]" : empty_string,
				Client->user,
				flags & DCC_DELETE      ? "Closed" :
				flags & DCC_ACTIVE      ? "Active" : 
				flags & DCC_MY_OFFER    ? "Waiting" :
				flags & DCC_THEIR_OFFER ? "Offered" : 
							  "Unknown",
				time_f, size, completed, filename);
		}
	}
	do_hook(DCC_LIST_LIST, "End * * * * * * *");
}


/*
 * Usage: /DCC RAW <filedesc> <host> [text]
 */
static	void	dcc_send_raw (char *args)
{
	char	*name;
	char	*host;

	if (!(name = next_arg(args, &args)))
	{
		say("No name specified for DCC RAW");
		return;
	}
	if (!(host = next_arg(args, &args)))
	{
		say("No hostname specified for DCC RAW");
		return;
	}
	dcc_message_transmit(DCC_RAW, name, host, args, args, 1, NULL);
}

/*
 * Usage: /DCC RENAME <nick> [<oldname>] <newname>
 */
static	void	dcc_rename (char *args)
{
	DCC_list	*dcc;
	char	*user;
	char	*oldf;
	char	*newf;
	char	*temp;
	int	type = DCC_FILEREAD;
	
	if (!(user = next_arg(args, &args)) || !(temp = next_arg(args, &args)))
	{
		say("You must specify a nick and new filename for DCC RENAME");
		return;
	}

	if ((newf = next_arg(args, &args)) != NULL)
		oldf = temp;
	else
	{
		newf = temp;
		oldf = NULL;
	}

	if (!my_strnicmp(user, "-CHAT", strlen(user)))
	{
		if (!oldf || !newf)
		{
		    say("You must specify a new nickname for DCC RENAME -CHAT");
		    return;
		}
		user = oldf;
		oldf = "chat";
		type = DCC_CHAT;
	}

	if ((dcc = dcc_searchlist(oldf, user, type, 0, NULL, -1)))
	{
		if (type == DCC_FILEREAD && (dcc->flags & DCC_ACTIVE))
		{
			say("Too late to rename that file");
			return;
		}

		if (type == DCC_FILEREAD)
		{
			say("File %s from %s renamed to %s",
				dcc->description ? dcc->description : "<any>",
				user, newf);
			malloc_strcpy(&dcc->description, newf);
		}
		else
		{
			say("DCC CHAT from %s changed to new nick %s",
				user, newf);
			malloc_strcpy(&dcc->user, newf);
		}
	}
	else
		say("%s has not yet offered you the file %s",
			user, oldf ? oldf : "<any>");
}


/*
 * Usage: /DCC REVERSE <type> <nick> <filename>
 */
static	void	dcc_reverse (char *args)
{
	DCC_list	*dcc;
	char	*stype;
	char	*user;
	char	*file;
	int	type;
	int	i;

	panic("Not implemented, so there.  Nyaa.");
	
	if (!(stype = next_arg(args, &args)) || !(user = next_arg(args, &args)))
	{
		say("You must specify a type and nickname for DCC REVERSE");
		return;
	}

	file = next_arg(args, &args);

	type = -1;
	for (i = 0; my_stricmp(dcc_types[i], stype); i++)
	{
		type = i + 1;
		break;
	}

	if (type == -1)
	{
		say("You must specify a valid type for DCC REVERSE");
		return;
	}

	if (!(dcc = dcc_searchlist(file, user, type, 0, NULL, 0)))
	{
		say("There are no pending %s connections from %s",
			dcc_types[i], user);
		return;
	}

	say("DCC %s from %s reversed direction", dcc_types[i], user);
}


/*
 * Usage: /DCC SEND <nick> <filename> [<filename>]
 */
static	void	dcc_filesend (char *args)
{
	char		*user,
			*fullname,
			*this_arg,
			FileBuf[BIG_BUFFER_SIZE+1];
	int		portnum = 0,
			filenames_parsed = 0;
	DCC_list	*Client;
	struct	stat	stat_buf;

	/*
	 * For sure, at least one argument is needed, the target
	 */
	if ((user = next_arg(args, &args)))
	{
	    while (args && *args)
	    {
		this_arg = new_next_arg(args, &args);

		/*
		 * Check to see if its a flag
		 */
		if (*this_arg == '-')
		{
			if (this_arg[1] == 'p')
			{
				if (args && *args)
				    portnum = my_atol(next_arg(args, &args));
			}
			continue;
		}

		/*
		 * Ok.  We have a filename they want to send.  Check to
		 * see what kind it is.
		 */

		/*
		 * Absolute pathnames are a can of corn.
		 */
		if (*this_arg == '/')
			strlcpy(FileBuf, this_arg, BIG_BUFFER_SIZE);
		/*
		 * Twiddle pathnames need to be expanded.
		 */
		else if (*this_arg == '~')
		{
			if (!(fullname = expand_twiddle(this_arg)))
			{
				say("Unable to expand %s", this_arg);
				return;
			}
			strlcpy(FileBuf, fullname, BIG_BUFFER_SIZE);
			new_free(&fullname);
		}
		/*
		 * Relative pathnames get the cwd tacked onto them.
		 */
		else
		{
			getcwd(FileBuf, BIG_BUFFER_SIZE);
			strmcat(FileBuf, "/", BIG_BUFFER_SIZE);
			strmcat(FileBuf, this_arg, BIG_BUFFER_SIZE);
		}

		/*
		 * Make a note that we've seen a filename
		 */
		filenames_parsed++;

		if (stat(FileBuf, &stat_buf))
		{
			say("The file %s is unaccessable (it doesn't exist or you dont have permission to access it.)", FileBuf);
			continue;
		}

		if (access(FileBuf, R_OK))
		{
			say("Cannot send %s because you dont have read permission", FileBuf);
			continue;
		}

		if (stat_buf.st_mode & S_IFDIR)
		{
			say("Cannot send %s because it is a directory", FileBuf);
			continue;
		}

		/* XXXXX filesize is a global XXXXX */
		filesize = stat_buf.st_size;
		Client = dcc_searchlist(FileBuf, user, DCC_FILEOFFER, 
					1, this_arg, -1);
		filesize = 0;

		if ((Client->flags & DCC_ACTIVE) ||
		    (Client->flags & DCC_MY_OFFER))
		{
			say("Sending a booster CTCP handshake for an existing "
				"DCC SEND:%s to %s", FileBuf, user);
			dcc_send_booster_ctcp(Client);
			continue;
		}

		Client->flags |= DCC_TWOCLIENTS;
		Client->local_port = portnum;
		dcc_open(Client);
	    } /* The WHILE */
	} /* The IF */

	if (!filenames_parsed)
		yell("Usage: /DCC SEND <[=]nick> <file> [<file> ...]");
}


/*
 * Usage: $listen(<port>)
 */
char	*dcc_raw_listen (unsigned short port)
{
	DCC_list	*Client;
	char		*PortName;

	if (port && port < 1025)
	{
		set_display_target(NULL, LOG_DCC);
		say("May not bind to a privileged port");
		reset_display_target();

		return m_strdup(empty_string);
	}

	PortName = LOCAL_COPY(ltoa(port));
	Client = dcc_searchlist("raw_listen", PortName, 
					DCC_RAW_LISTEN, 1, NULL, -1);

	if (Client->flags & DCC_ACTIVE)
	{
		set_display_target(NULL, LOG_DCC);
		say("A previous DCC RAW_LISTEN on %s exists", PortName);
		reset_display_target();

		return m_strdup(empty_string);
	}

	if ((Client->peer = connect_by_number(NULL, &port, 
				SERVICE_SERVER, PROTOCOL_TCP, NULL)) < 0)
	{
		Client->flags |= DCC_DELETE; 
		set_display_target(NULL, LOG_DCC);
		say("Couldn't establish listening socket: [%d] %s", 
				Client->peer, strerror(errno));
		reset_display_target();

		Client->peer = -1;
		return m_strdup(empty_string);
	}

	accept_callback(Client->peer, dcc_raw_accept);
	get_time(&Client->starttime);
	Client->local_port = port;
	Client->flags |= DCC_ACTIVE;
	Client->user = m_strdup(ltoa(Client->local_port));
	return m_strdup(Client->user);
}


/*
 * Usage: $connect(<hostname> <portnum>)
 */
char	*dcc_raw_connect (char *host, u_short port)
{
	DCC_list	*Client;
	IA		addr;
	char *		bogus;

	/*
	 * Look up the internet address for the specified hostname.
	 */
	addr = inet_anyton(host);
	if (ADDRESS(addr) == INADDR_NONE)
	{
		set_display_target(NULL, LOG_DCC);
		say("Unknown host: %s", host);
		reset_display_target();
		return m_strdup(empty_string);
	}

	/*
	 * Create a new DCC session for this connection.
	 */
	bogus = LOCAL_COPY(ltoa(port));
	Client = dcc_searchlist(host, bogus, DCC_RAW, 1, NULL, -1);

	/*
	 * Don't permit duplicate connections since the list can't handle that.
	 * XXX I don't think this can happen.
	 */
	if (Client->flags & DCC_ACTIVE)
	{
		set_display_target(NULL, LOG_DCC);
		say("A previous DCC RAW to %s on %s exists", host, bogus);
		reset_display_target();

		return m_strdup(empty_string);
	}

	Client->flags = DCC_THEIR_OFFER | DCC_RAW;
	Client->remote_addr = addr;
	Client->remote_port = htons(port);
	if (!dcc_open(Client))
		return m_strdup(empty_string);

	Client->user = m_strdup(ltoa(Client->peer));

	set_display_target(NULL, LOG_DCC);
	if (do_hook(DCC_RAW_LIST, "%s %s E %d", 
			Client->user, host, port))
            if (do_hook(DCC_CONNECT_LIST,"%s RAW %s %d",
			Client->user, host, port))
		say("DCC RAW connection to %s on %hd via %d established", 
			host, port, Client->peer);
	reset_display_target();

	return m_strdup(Client->user);
}





/*
 *
 * All the rest of this file is dedicated to automatic replies generated
 * by the client in response to external stimuli from DCC connections.
 *
 */



/*
 * When we recieve a CTCP DCC from a remote user, it comes here to
 * be handled.
 */
void	register_dcc_offer (char *user, char *type, char *description, char *address, char *port, char *size, char *extra)
{
	DCC_list *	Client;
	int		CType;
	char *		c;
	int		do_auto = 0;	/* used in dcc chat collisions */

	if (x_debug & DEBUG_DCC_SEARCH)
		yell("register_dcc_offer: [%s|%s|%s|%s|%s|%s|%s]", 
			user, type, description, address, port, size, extra);

	/*
	 * If the remote peer was sloppy and sent over a full pathname, then
	 * just extract the interesting part of it.  If they are sending a
	 * dot-file, then change that dot to a _ ! ;-)
	 */
	if ((c = strrchr(description, '/')))
		description = c + 1;
	if (*description == '.')
		*description = '_';

	/*
	 * Extract the file size from the handshake.  XXX This is a global
	 * variable referenced by dcc_searchlist, though it should be handled
	 * some other way.
	 */
	if (size && *size)
		filesize = my_atol(size);

	/*
	 * Determine what kind of DCC request this.
	 */
	if (!my_stricmp(type, "CHAT"))
		CType = DCC_CHAT;
	else if (!my_stricmp(type, "SEND"))
		CType = DCC_FILEREAD;
#ifdef MIRC_BROKEN_DCC_RESUME
	else if (!my_stricmp(type, "RESUME"))
	{
		/* 
		 * Dont be deceieved by the arguments we're passing it.
		 * The arguments are "out of order" because MIRC doesnt
		 * send them in the traditional order.  Ugh.
		 */
		dcc_getfile_resume_demanded(user, description, address, port);
		return;
	}
	else if (!my_stricmp(type, "ACCEPT"))
	{
		/*
		 * See the comment above.
		 */
		dcc_getfile_resume_start (user, description, address, port);
		return;
	}
#endif
        else
        {
                say("Unknown DCC %s (%s) received from %s", 
				type, description, user);
                return;
        }

	/*
	 * Create a new dcc object for this session.  If there is already
	 * one open, we'll find out soon enough.
	 */
	Client = dcc_searchlist(description, user, CType, 1, NULL, -1);

	/*
	 * If I have sent out an offer that matches this exactly, then
	 * we have a collision.  Bump off the old one and punt.
	 */
	if (Client->flags & DCC_MY_OFFER)
	{
		/* Mark the old entry as revoked */
		Client->flags |= DCC_DELETE;

		/* 
		 * If it is a DCC CHAT collision, then create a new
		 * session on their terms, and then continue as normal.
		 */
		if (DCC_CHAT == CType)
		{
			Client = dcc_searchlist(description, user, CType, 
						1, NULL, -1);
			do_auto = 1;
		}

		/*
		 * Otherwise, just return an error to the other user.
		 * XXX - This should be a reject
		 * XXX - This seems pointless.
		 */
		else
		{
			say("DCC %s collision for %s:%s", 
				type, user, description);
			send_ctcp(CTCP_NOTICE, user, CTCP_DCC,
				"DCC %s collision occured while "
				"connecting to %s (%s)", 
				type, 
				get_server_nickname(from_server), description);

			return;
		}
	}

	/*
	 * If the connection is current live and well, then punt this
	 * request as it makes no sense.
	 */
	if (Client->flags & DCC_ACTIVE)
	{
		say("Received DCC %s request from %s while previous "
		    "session still active", type, user);
		return;
	}

	/*
	 * Ok.  Now we are in business.  Stash away the checksum and 
	 * reset that bogus filesize global variable.
	 */
	Client->flags |= DCC_THEIR_OFFER;

	get_time(&Client->lasttime);
	get_time(&Client->starttime);

	/*
	 * 'x' is reserved for future timestamp extensions,
	 * start ignoring it now.
	 */
	if (extra && *extra && *extra != 'x')
		Client->cksum = m_strdup(extra);
	filesize = 0;


	/* EXTRACT THE IPv4 ADDRESS FROM THE HANDSHAKE */
	/*
	 * Get the hostname from the handshake.  We support the 
	 * traditional network-order 32 bit hostnames as well as 
	 * the CTCP/2 mandated ascii A.B.C.D hostnames (which are not 
	 * used by anybody at the time i wrote this comment.)
	 */
	if (strchr(address, '.'))
	{
		say("DCC %s (%s) request from %s had an invalid address [%s]", 
			type, description, user, address);
		return;
	}
	else
	{
		ADDRESS(Client->remote_addr) = htonl(strtoul(address, NULL, 10));
		if (ADDRESS(Client->remote_addr) == 0)
		{
			Client->flags |= DCC_DELETE;
			yell("### DCC handshake from %s ignored becuase it "
				"had an null address", user);
			return;
		}
	}

	/* VERIFY THE IPv4 ADDRESS IF REQUESTED */
	/* 
	 * This compares the IP address in the handshake against the 
	 * IP address of the hostname contained in the PRIVMSG protocol
	 * message.  This is not the default because if the hostname in 
	 * the PRIVMSG is bogus, it could block up to one second.  On 
	 * networks where hostname spoofing is not a problem, this check 
	 * is not really a big deal.  But where it is a problem, this can
	 * be used by floodbots as a denial of service attack.
	 */
	if (get_int_var(DCC_CHECK_HANDSHAKE_ADDR_VAR))
	{
	    char	*offer_userhost;
	    char	*offer_host = NULL;
	    IA		offer_address;

	    offer_userhost = LOCAL_COPY(FromUserHost);
	    if ((offer_host = strchr(offer_userhost, '@')))
	    {
		*offer_host++ = 0;
		offer_address = inet_anyton(offer_host);

		if (ADDRESS(offer_address) == INADDR_NONE)
		{
		    yell("### WARNING! ###");
		    yell("### Address in the PRIVMSG [%s] is invalid", 
				offer_host);
		}
		else if (ADDRESS(offer_address) != ADDRESS(Client->remote_addr))
		{
		    yell("### WARNING! ###");
		    yell("### Address in the PRIVMSG [%s (%s)] does not match",
				offer_host, inet_ntoa(offer_address));
		    yell("### Address in the handshake [(%d)]",
				ADDRESS(Client->remote_addr));
		    yell("### Make sure you know this request is safe before");
		    yell("### you accept it!");
		}
	    }
	}

	/* EXTRACT THE PORT INFORMATION FROM THE HANDSHAKE */
	/*
	 * Extract the port.  Do some sanity checking.
	 */
	Client->remote_port = htons((u_short)strtoul(port, NULL, 10));
	if (Client->remote_port < 1024)
	{
		Client->flags |= DCC_DELETE;
		say("DCC %s (%s) request from %s rejected because it "
			"specified a reserved port number", 
				type, description, user);
		return;
	}


	/*
	 * Now tell the user about this handshake.  This is unfortunately
	 * long and complicated, but it has to be done.
	 */

	/*
	 * If this is a DCC CHAT collision, then we have already determined
	 * to automatically go forward with this connection.  Do that here.
	 * (We do this here rather than above, because we still want to make
	 * sure that the port and hostnames are valid, and they had not 
	 * yet been checked.)
	 */
  	if (do_auto)
  	{
                if (do_hook(DCC_CONNECT_LIST,"%s CHAT", user))
			say("DCC CHAT already requested by %s, connecting...",
				user);
  		dcc_chat(user);
		return;
	}

       /* 
	* DCC SEND and CHAT have different arguments, so they can't
	* very well use the exact same hooked data.  Both now are
	* identical for $0-4, and SEND adds filename/size in $5-6 
	*/
	if (DCC_TYPE(Client->flags) == DCC_FILEREAD)
	{
		if (!do_hook(DCC_REQUEST_LIST,
				"%s %s %s %s %d %s %ld",
				user, type, description,
				inet_ntoa(Client->remote_addr),
				ntohs(Client->remote_port),
				Client->description,
				Client->filesize))
			return;
	}
	else
	{
		if (!do_hook(DCC_REQUEST_LIST,
				"%s %s %s %s %d",
				user, type, description,
				inet_ntoa(Client->remote_addr),
				ntohs(Client->remote_port)))
			return;
	}

	/*
	 * If someone is trying to send us a file, see if the file
	 * exists and inform the user as to their choices in the matter.
	 */
	if (DCC_TYPE(Client->flags) == DCC_FILEREAD)
	{
		struct stat statit;
		if (stat(Client->description, &statit) == 0)
		{
		    if (statit.st_size < Client->filesize)
		    {
			say("WARNING: File [%s] exists but is smaller "
				"then the file offered.", 
					Client->description);
			say("Use /DCC CLOSE GET %s %s        to "
				"not get the file.", 
					user, Client->description);
#ifdef MIRC_BROKEN_DCC_RESUME
			say("Use /DCC RESUME %s %s           to "
				"continue the copy where it left off.",
					 user, Client->description);
#endif
			say("Use /DCC RENAME %s %s newname   to "
				"save it to a different filename", 
					user, Client->description);
			say("Use /DCC GET %s %s              to "
				"overwrite the existing file.", 
					user, Client->description);
		    }

		    else if (statit.st_size == Client->filesize)
		    {
			say("WARNING: File [%s] exists, and it is "
				"the same size.", 
					Client->description);
			say("Use /DCC CLOSE GET %s %s        to "
				"not get the file.", 
					user, Client->description);
			say("Use /DCC RENAME %s %s           to "
				"get the file anyways under a "
				"different name.", 
					user, Client->description);
			say("Use /DCC GET %s %s              to "
				"overwrite the existing file.", 
					user, Client->description);
		    }

		    else 	/* Bigger than */
		    {
			say("WARNING: File [%s] already exists.", 
				Client->description);
			say("Use /DCC CLOSE GET %s %s        to "
				"not get the file.", 
					user, Client->description);
			say("Use /DCC RENAME %s %s           to "
				"get the file anyways under a "
				"different name.", 
					user, Client->description);
			say("Use /DCC GET %s %s              to "
				"overwrite the existing file.", 
					user, Client->description);
		    }
		}
	}

	/*
	 * If we have gotten this far, then we know that the user has not
	 * bitten at the request, and they have been warned if neccesary.
	 * So here we tell them that a connection has been offered.
	 */
	/* Thanks, Tychy! (lherron@imageek.york.cuny.edu) */
	if (DCC_TYPE(Client->flags) == DCC_FILEREAD)
		say("DCC %s (%s %ld) request received from %s!%s [%s:%d]",
			type, description, Client->filesize, user, 
			FromUserHost, inet_ntoa(Client->remote_addr), 
			ntohs(Client->remote_port));
	else
	        say("DCC %s (%s) request received from %s!%s [%s:%d]", 
			type, description, user, FromUserHost, 
			inet_ntoa(Client->remote_addr), 
			ntohs(Client->remote_port));

	return;
}


/**************************************************************************/
/**************************************************************************/
/*
 * Check all DCCs for data, and if they have any, perform whatever
 * actions are required.
 */
void	dcc_check (void)
{
	DCC_list	*Client;

	set_display_target(NULL, LOG_DCC);
	for (Client = ClientList; Client; Client = Client->next)
	{
		/* 
		 * There are some ways that this can happen.
		 * Rather than track them all down and "fix" them,
		 * I'll just conceed that this is possible and skip them.
		 */
		if (Client->flags & DCC_DELETE)
		{
			dcc_add_deadclient(Client);
			continue;
		}

		/*
		 * Enforce any timeouts
		 */
		else if (Client->flags & (DCC_MY_OFFER | DCC_THEIR_OFFER) &&
				dcc_timeout >= 0 &&
				(now - Client->lasttime.tv_sec > dcc_timeout))
		{
			if (do_hook(DCC_LOST_LIST,"%s %s %s IDLE TIME EXCEEDED",
				Client->user, 
				dcc_types[Client->flags & DCC_TYPES],
				Client->description ? 
				 Client->description : "<any>"))
			    say("Auto-rejecting a dcc after [%ld] seconds", 
				now - Client->lasttime.tv_sec);

			/*
			 * This is safe to do after, since the connection
			 * is still open, and the user might want to send
			 * something to the other peer here, and there is no
			 * good reason not to let them.
			 */
			Client->flags |= DCC_DELETE;
		}

		if (Client->flags & DCC_DELETE)
			dcc_add_deadclient(Client);
	}

	/* Garbage collect anything that is left */
	if (deadlist)
		dcc_really_erase();

	reset_display_target();
}



/**************************************************************************/
static	int	dcc_chat_accept (int old, int new, char *caddr)
{
	ISA	remaddr;
	DCC_list *Client;
	int	rsize = sizeof(ISA);

	for (Client = ClientList; Client; Client = Client->next)
		if (Client->peer == old)
			break;
	if (!Client)
		panic("Invalid dcc_chat_accept callback");

	new_close(Client->peer);
	Client->peer = new;
	line_buffering(Client->peer, &dcc_chat_discipline);

	Client->flags &= ~DCC_MY_OFFER;
	Client->flags |= DCC_ACTIVE;
	get_time(&Client->starttime);

	if (getpeername(Client->peer, (SA *)&remaddr, &rsize))
	{
		yell("Hello?  Beuller?");
		return 0;
	}
	Client->remote = remaddr;

	if (do_hook(DCC_CONNECT_LIST, "%s CHAT %s %d", 
			Client->user,
			inet_ntoa(remaddr.sin_addr), 
			ntohs(remaddr.sin_port)))
	    say("DCC chat connection to %s[%s:%d] established", 
			Client->user, inet_ntoa(remaddr.sin_addr), 
			ntohs(remaddr.sin_port));

	get_time(&Client->starttime);
	return 1;
}

static	int	dcc_chat_error (int fd)
{
	char *	real_tmp;
	DCC_list *Client;

	for (Client = ClientList; Client; Client = Client->next)
		if (Client->peer == fd)
			break;
	if (!Client)
		panic("Invalid dcc_chat_error callback");

	if (io_errno == -1)
		real_tmp = "Remote End Closed Connection";
	else
		real_tmp = strerror(io_errno);

	Client->flags |= DCC_DELETE;

	if (do_hook(DCC_LOST_LIST, "%s CHAT %s", Client->user, real_tmp))
		say("DCC CHAT connection to %s lost [%s]", 
				Client->user, real_tmp);

	Client->peer = -1;
	return 0;
}

static	int	dcc_chat_reader (int fd, int bytesread, char *tmp)
{
	char 	uh	[80];
	char 	eq_nick	[80];
	DCC_list *Client;

	for (Client = ClientList; Client; Client = Client->next)
		if (Client->peer == fd)
			break;
	if (!Client)
		panic("Invalid dcc_chat_error callback");

	chomp(tmp);
	my_decrypt(tmp, strlen(tmp), Client->encrypt);
	Client->bytes_read += bytesread;

	if (x_debug & DEBUG_INBOUND) 
		yell("%s", tmp);

#define CTCP_MESSAGE "CTCP_MESSAGE "
#define CTCP_MESSAGE_LEN strlen(CTCP_MESSAGE)
#define CTCP_REPLY "CTCP_REPLY "
#define CTCP_REPLY_LEN strlen(CTCP_REPLY)

	if (*tmp == CTCP_DELIM_CHAR)
	{
		snprintf(eq_nick, 80, "=%s", Client->user);
		snprintf(uh, 80, "%s@%s", "unknown",
					inet_ntoa(Client->remote_addr));
		FromUserHost = uh;
		tmp = do_ctcp(eq_nick, nickname, LOCAL_COPY(tmp));
		FromUserHost = empty_string;
	}
	if (!strncmp(tmp, CTCP_MESSAGE, CTCP_MESSAGE_LEN))
	{
		snprintf(eq_nick, 80, "=%s", Client->user);
		snprintf(uh, 80, "%s@%s", 
				"unknown", inet_ntoa(Client->remote_addr));

		FromUserHost = uh;
		tmp = do_ctcp(eq_nick, nickname, 
				LOCAL_COPY(tmp + CTCP_MESSAGE_LEN));
		FromUserHost = empty_string;
	}
	else if (!strncmp(tmp, CTCP_REPLY, CTCP_REPLY_LEN) || 
			*tmp == CTCP_DELIM_CHAR)
	{
		snprintf(eq_nick, 80, "=%s", Client->user);
		snprintf(uh, 80, "%s@%s", 
				"unknown", inet_ntoa(Client->remote_addr));

		FromUserHost = uh;
		if (*tmp == CTCP_DELIM_CHAR)
			tmp = do_notice_ctcp(eq_nick, nickname, 
				LOCAL_COPY(tmp));
		else
			tmp = do_notice_ctcp(eq_nick, nickname, 
				LOCAL_COPY(tmp + CTCP_REPLY_LEN));
		FromUserHost = empty_string;
	}

	if (!*tmp)
		return 0;

	set_display_target(Client->user, LOG_DCC);
	if (do_hook(DCC_CHAT_LIST, "%s %s", Client->user, tmp))
		put_it("=%s= %s", Client->user, tmp);
	reset_display_target();

	return 0;
}

/**************************************************************************/
static	int	dcc_raw_accept (int fd, int new_fd, char *addr)
{
	ISA *		remaddr;
const	char *		name;
	char		fd_name[10];
	DCC_list	*Client;
	DCC_list	*old;

	for (old = ClientList; old; old = old->next)
		if (old->peer == fd)
			break;
	if (!old)
		panic("Invalid DCC RAW callback");


	remaddr = (ISA *)addr;
	name = inet_ntoany(remaddr->sin_addr);

	strlcpy(fd_name, ltoa(new_fd), 10);
	Client          = dcc_searchlist(name, fd_name, DCC_RAW, 1, NULL, 0);
	Client->peer        = new_fd;
	Client->remote	    = *remaddr;
	Client->flags      |= DCC_ACTIVE;
	Client->bytes_read  = Client->bytes_sent = 0L;
	get_time(&Client->starttime);
	line_buffering(Client->peer, &dcc_raw_discipline);

	if (do_hook(DCC_RAW_LIST, "%s %s N %d", Client->user,
						Client->description,
						Client->local_port))
        if (do_hook(DCC_CONNECT_LIST,"%s RAW %s %d", Client->user,
                                                     Client->description,
                                                     old->local_port))
		say ("DCC RAW connection to %s on %s via %d established",
					Client->description,
					Client->user,
					old->local_port);

	return 1;
}

static	int	dcc_raw_error (int fd)
{
	DCC_list	*client;

	for (client = ClientList; client; client = client->next)
		if (client->peer == fd)
			break;
	if (!client)
		panic("Invalid DCC RAW ERROR callback");

	client->flags |= DCC_DELETE;

	if (do_hook(DCC_RAW_LIST, "%s %s C",
			client->user, client->description))
	if (do_hook(DCC_LOST_LIST,"%s RAW %s", 
			client->user, client->description))
		say("DCC RAW connection to %s on %s lost",
			client->user, client->description);

	client->peer = -1;
	return 0;
}

static	int	dcc_raw_reader (int fd, int bytesread, char *tmp)
{
	DCC_list	*client;

	for (client = ClientList; client; client = client->next)
		if (client->peer == fd)
			break;
	if (!client)
		panic("Invalid DCC RAW ERROR callback");

	chomp(tmp);
	client->bytes_read += bytesread;

	from_server = primary_server;	/* Colten asked for it */
	if (do_hook(DCC_RAW_LIST, "%s %s D %s",
			client->user, client->description, tmp))
		say("Raw data on %s from %s: %s",
			client->user, client->description, tmp);

	return 1;
}

/**************************************************************************/
static	int	dcc_send_accept (int old_fd, int new_fd, char *caddr)
{
	int		size;
	DCC_list *	Client;
	ISA		remaddr;
	int		rsize = sizeof(remaddr);

	for (Client = ClientList; Client; Client = Client->next)
		if (Client->peer == old_fd)
			break;
	if (!Client)
		panic("Invalid DCC RAW ERROR callback");

	new_close(Client->peer);
	Client->peer = new_fd;
	Client->flags &= ~DCC_MY_OFFER;
	Client->flags |= DCC_ACTIVE;
	if (getpeername(Client->peer, (SA *)&remaddr, &rsize))
	{
		yell("Hello?  Beuller?");
		return 0;
	}
	Client->remote = remaddr;
	get_time(&Client->starttime);

	no_buffering_readonly(Client->peer, &dcc_send_discipline);

#ifdef HAVE_SO_SNDLOWAT
	/*
	 * Give a hint to the OS how many bytes we need to send
	 * for each write().  Some OS's, like linux, can be bothered
	 * to #define it, but not to implement it.  Sheesh.  How bogus.
	 */
	size = DCC_BLOCK_SIZE;
	if (setsockopt(Client->peer, SOL_SOCKET, SO_SNDLOWAT, 
				&size, sizeof(size)) < 0)
		say("setsockopt failed: %s", strerror(errno));
#endif

	/*
	 * Tell the user
	 */
	set_display_target(NULL, LOG_DCC);
	if (do_hook(DCC_CONNECT_LIST, "%s SEND %s %d %s %ld",
		Client->user, inet_ntoa(Client->remote_addr),
		ntohs(Client->remote_port), Client->description,
		Client->filesize))
	    say("DCC SEND connection to %s[%s (%s):%d] established", 
		Client->user, inet_ntoany(Client->remote_addr),
		inet_ntoa(Client->remote_addr), ntohs(Client->remote_port));
	reset_display_target();

	/*
	 * Open up the file to be sent
	 */
	if ((Client->file = open(Client->description, O_RDONLY)) == -1)
	{
		Client->flags |= DCC_DELETE;

		set_display_target(NULL, LOG_DCC);
		say("Unable to open %s: %s", Client->description,
			errno ? strerror(errno) : "Unknown Host");
		reset_display_target();

		return 1;
	}

	/*
	 * Set up any DCC RESUME as needed
	 */
	if (Client->bytes_sent)
	{
		set_display_target(NULL, LOG_DCC);
		if (x_debug & DEBUG_DCC_XMIT)
			yell("Resuming at address (%ld)", 
				(long)Client->bytes_sent);
		reset_display_target();

		lseek(Client->file, Client->bytes_sent, SEEK_SET);
	}

	return dcc_send_packets(Client);
}


static	int	dcc_send_error (int fd)
{
	DCC_list *		Client;

	for (Client = ClientList; Client; Client = Client->next)
		if (Client->peer == fd)
			break;
	if (!Client)
		panic("Invalid DCC SEND ERROR callback");

	Client->flags |= DCC_DELETE;
	Client->peer = -1;

	set_display_target(NULL, LOG_DCC);
	if (do_hook(DCC_LOST_LIST,"%s SEND %s CONNECTION LOST",
		Client->user, Client->description))
	    say("DCC SEND:%s connection to %s lost: %s",
		Client->description, Client->user, strerror(io_errno));
	reset_display_target();

	return 0;
}


static	int	dcc_send_reader (int fd, int len, char *buf)
{
	DCC_list *		Client;
	u_32int_t		bytesrecvd = 0;

	for (Client = ClientList; Client; Client = Client->next)
		if (Client->peer == fd)
			break;
	if (!Client)
		panic("Invalid DCC SEND READER callback");

	if (x_debug & DEBUG_DCC_XMIT)
	{
		set_display_target(NULL, LOG_DCC);
		yell("Reading a packet from [%s:%s(%ld)]", 
			Client->user, 
			Client->othername, 
			Client->packets_ack);
		reset_display_target();
	}

	/*
	 * It is important to note here that the ACK must always
	 * be exactly four bytes.  Never more, never less.
	 */
	while (len > 0)
	{
#if 0
	    if (len != sizeof(u_32int_t))
		return dcc_send_error(fd);
#endif
	    bytesrecvd = ntohl(*(u_32int_t *)buf);
	    buf += sizeof(u_32int_t);
	    len -= sizeof(u_32int_t);

	    /*
	     * Check to see if we need to move the sliding window up
	     */
	    if (bytesrecvd >= (Client->packets_ack + 1) * DCC_BLOCK_SIZE)
	    {
		if (x_debug & DEBUG_DCC_XMIT)
		{
			set_display_target(NULL, LOG_DCC);
			yell("Packet #%ld ACKed", Client->packets_ack);
			reset_display_target();
		}
		Client->packets_ack++;
		Client->packets_outstanding--;
	    }

	    if (bytesrecvd > Client->bytes_sent)
	    {
set_display_target(NULL, LOG_DCC);
yell("### WARNING!  The other peer claims to have recieved more bytes than");
yell("### I have actually sent so far.  Please report this to ");
yell("### jnelson@acronet.net or #epic on EFNet right away.");
yell("### Ask the person who you sent the file to look for garbage at the");
yell("### end of the file you just sent them.  Please inclose that ");
yell("### information as well as the following:");
yell("###");
yell("###    bytesrecvd [%ld]        Client->bytes_sent [%ld]", 
			ntohl(bytesrecvd), (long)Client->bytes_sent);
yell("###    Client->filesize [%ld]  Client->packets_ack [%ld]",
			Client->filesize, Client->packets_ack);
yell("###    Client->packets_outstanding [%ld]",
			Client->packets_outstanding);
reset_display_target();

		/* And just cope with it to avoid whining */
		Client->bytes_sent = bytesrecvd;
	    }
	}

	/*
	 * If we've sent the whole file already...
	 */
	if (Client->bytes_sent >= Client->filesize)
	{
		/*
		 * If theyve ACKed the last packet, we close 
		 * the connection.
		 */
		if (bytesrecvd >= Client->filesize)
			DCC_close_filesend(Client, "SEND");

		/*
		 * Either way there's nothing more to do.
		 */
		return 0;
	}

	return dcc_send_packets(Client);
}

static int	dcc_send_packets (DCC_list *Client)
{
	int		maxpackets;
	fd_set		fds;
	Timeval 	to;
	char		tmp[DCC_BLOCK_SIZE+1];
	int		bytesread;

	/*
	 * In either case, we need to send some more data to the peer.
	 * We use a nonblocking sliding window algorithm.  We send as many
	 * packets as we can *without blocking*, up to but never more than
	 * the value of /SET DCC_SLIDING_WINDOW.  Whenever we recieve some
	 * stimulus (like from an ACK) we re-fill the window.  We always do
	 * a select() before we write() to make sure that it wont block.
	 */
	FD_ZERO(&fds);
	if ((maxpackets = get_int_var(DCC_SLIDING_WINDOW_VAR)) < 1)
		maxpackets = 1;		/* Sanity */

	while (Client->packets_outstanding < maxpackets)
	{
		/*
		 * Check to make ure the write wont block.
		 */
		FD_SET(Client->peer, &fds);
		to.tv_sec = 0;
		to.tv_usec = 0;
		if (select(Client->peer + 1, NULL, &fds, NULL, &to) <= 0)
			break;

		/*
		 * Grab some more file.  If this chokes, dont sweat it.
		 */
		if ((bytesread = read(Client->file, tmp, DCC_BLOCK_SIZE)) <= 0)
			break;


		/*
		 * Bug the user
		 */
		if (x_debug & DEBUG_DCC_XMIT)
		{
		    set_display_target(NULL, LOG_DCC);
		    yell("Sending packet [%s [%s] (packet %ld) (%d bytes)]",
			Client->user, 
			Client->othername, 
			Client->packets_transfer, 
			bytesread);
		    reset_display_target();
		}

		/*
		 * Attempt to write the file.  If it chokes, whine.
		 */
		if (write(Client->peer, tmp, bytesread) < bytesread)
		{
			Client->flags |= DCC_DELETE;

			set_display_target(NULL, LOG_DCC);
			say("Outbound write() failed: %s", 
				errno ? strerror(errno) : "I'm not sure why");
			reset_display_target();

			return 1;
		}

		Client->packets_outstanding++;
		Client->packets_transfer++;
		Client->bytes_sent += bytesread;

		/*
		 * Update the status bar
		 */
		if (Client->filesize)
		{
		    update_transfer_buffer("(to %10s: %d of %d: %d%%)",
			Client->user, 
			Client->packets_transfer, 
			Client->packets_total, 
			(Client->bytes_sent * 100 / Client->filesize));
		}
		else
		{
		    update_transfer_buffer("(to %10s: %d kbytes     )",
			Client->user, 
			Client->bytes_sent / 1024);
		}
		update_all_status();
	}

	return 0;
}


/****************************************************************************/
static int	dcc_get_error (int fd)
{
	DCC_list *		Client;

	for (Client = ClientList; Client; Client = Client->next)
		if (Client->peer == fd)
			break;
	if (!Client)
		panic("Invalid DCC GET ERROR callback");

	if (Client->bytes_read < Client->filesize)
	{
		set_display_target(NULL, LOG_DCC);
		say("DCC GET to %s lost -- %s", Client->user,
				"Remote peer closed connection");
		reset_display_target();
	}

	Client->peer = -1;
	DCC_close_filesend(Client, "GET");
	return 0;
}

static	int	dcc_get_reader (int fd, int bytesread, char *tmp)
{
	u_32int_t	bytestemp;
	DCC_list *	Client;

	for (Client = ClientList; Client; Client = Client->next)
		if (Client->peer == fd)
			break;
	if (!Client)
		panic("Invalid DCC GET callback");


	my_decrypt(tmp, bytesread, Client->encrypt);
	if ((write(Client->file, tmp, bytesread)) == -1)
	{
		Client->flags |= DCC_DELETE;

		set_display_target(NULL, LOG_DCC);
		say("Write to local file failed.  Giving up.");
		reset_display_target();

		return 1;
	}


	Client->bytes_read += bytesread;
	bytestemp = htonl(Client->bytes_read);
	if (write(Client->peer, (char *)&bytestemp, sizeof(u_32int_t)) <= 0)
	{
		Client->flags |= DCC_DELETE;

		set_display_target(NULL, LOG_DCC);
		yell("### Write to remote peer failed.  Giving up.");
		reset_display_target();

		return 1;
	}

	Client->packets_transfer = Client->bytes_read / DCC_BLOCK_SIZE;

	/* 
	 * Protect against possible attacks by people who advertise one
	 * file size and then plug you into an infinite source of data.
	 * We only accept the number of bytes advertized.
	 */
	if ((DCC_TYPE(Client->flags) == DCC_FILEREAD) &&
			(Client->bytes_read > Client->filesize))
	{
		Client->flags |= DCC_DELETE;

		set_display_target(NULL, LOG_DCC);
		yell("### DCC GET WARNING: incoming file is "
			"larger then the handshake said");
		yell("### DCC GET: Closing connection");
		reset_display_target();

		return 1;
	}

	if ((DCC_TYPE(Client->flags) == DCC_FILEOFFER) ||
	    (DCC_TYPE(Client->flags) == DCC_FILEREAD))
	{
		if (Client->filesize)
		    update_transfer_buffer("(%10s: %d of %d: %d%%)", 
			Client->user, 
			Client->packets_transfer, 
			Client->packets_total, 
			Client->bytes_read * 100 / Client->filesize);
		else
		    update_transfer_buffer("(%10s %d packets: %dK)", 
			Client->user, 
			Client->packets_transfer, 
			Client->bytes_read / 1024);

		update_all_status();
	}
	return 0;
}


/*************************************************************************/
/*
 * This is a callback.  When we want to do a CTCP DCC REJECT, we do
 * a WHOIS to make sure theyre still on irc, no sense sending it to
 * nobody in particular.  When this gets called back, that means the
 * peer is indeed on irc, so we send them the REJECT.
 *
 * -- Changed this to be quasi-reentrant. yuck.
 */
static	void 	output_reject_ctcp (char *original, char *received)
{
	char	*nickname_requested;
	char	*nickname_recieved;
	char	*type;
	char	*description;

	/*
	 * XXX This is, of course, a monsterous hack.
	 */
	nickname_requested 	= next_arg(original, &original);
	type 			= next_arg(original, &original);
	description 		= next_arg(original, &original);
	nickname_recieved 	= next_arg(received, &received);

	if (nickname_recieved && *nickname_recieved)
		send_ctcp(CTCP_NOTICE, nickname_recieved, CTCP_DCC,
				"REJECT %s %s", type, description);
}



/*
 * This is called when someone sends you a CTCP DCC REJECT.
 */
void 	dcc_reject (char *from, char *type, char *args)
{
	DCC_list	*Client;
	char		*description;
	int		CType;

	for (CType = 0; dcc_types[CType] != NULL; CType++)
		if (!my_stricmp(type, dcc_types[CType]))
			break;

	if (!dcc_types[CType])
		return;

	description = next_arg(args, &args);

	if ((Client = dcc_searchlist(description, from, CType, 0, description, -1)))
	{
		Client->flags |= DCC_REJECTED;
		Client->flags |= DCC_DELETE; 

		set_display_target(NULL, LOG_DCC);
                if (do_hook(DCC_LOST_LIST,"%s %s %s REJECTED", from, type,
                        description ? description : "<any>"))
		    say("DCC %s:%s rejected by %s: closing", type,
			description ? description : "<any>", from);
		reset_display_target();
	}
}


static void 	DCC_close_filesend (DCC_list *Client, char *info)
{
	char	lame_ultrix[10];	/* should be plenty */
	char	lame_ultrix2[10];
	char	lame_ultrix3[10];
	double 	xtime, xfer;

	xtime = time_diff(Client->starttime, get_time(NULL));
	xfer = (double)(Client->bytes_sent ? Client->bytes_sent
					   : Client->bytes_read);
	sprintf(lame_ultrix, "%2.4g", (xfer / 1024.0 / xtime));

	/* Cant pass %g to put_it (lame ultrix/dgux), fix suggested by sheik. */
	if (xfer <= 0)
		xfer = 1;
	sprintf(lame_ultrix2, "%2.4g", xfer / 1024.0);

	if (xtime <= 0)
		xtime = 1;
	sprintf(lame_ultrix3, "%2.6g", xtime);

	Client->flags |= DCC_DELETE;

	set_display_target(NULL, LOG_DCC);
	if (do_hook(DCC_LOST_LIST,"%s %s %s %s TRANSFER COMPLETE",
		Client->user, info, Client->description, lame_ultrix))
	    say("DCC %s:%s [%skb] with %s completed in %s sec (%s kb/sec)",
		info, Client->description, lame_ultrix2, Client->user, 
		lame_ultrix3, lame_ultrix);
	reset_display_target();
}



/* 
 * Looks for the dcc transfer that is "current" (last recieved data)
 * and returns information for it
 */
char *	DCC_get_current_transfer (void)
{
	return DCC_current_transfer_buffer;
}


static void 	update_transfer_buffer (char *format, ...)
{
	if (format)
	{
		va_list args;
		va_start(args, format);
		vsnprintf(DCC_current_transfer_buffer, 255, format, args);
		va_end(args);
	}
	else
		*DCC_current_transfer_buffer = 0;

}

static        char *  dcc_urlencode (const char *s)
{
	const char *p1;
	char *str, *p2;

	str = m_strdup(s);

	for (p1 = s, p2 = str; *p1; p1++)
	{
		if (*p1 == '\\')
			continue;
		*p2++ = *p1;
	}

	*p2 = 0;
	return urlencode(str);
}

static        char *  dcc_urldecode (const char *s)
{
	char *str, *p1;

	str = m_strdup(s);
	urldecode(str);

	for (p1 = str; *p1; p1++)
	{
		if (*p1 != '.')
			break;
		*p1 = '_';
	}

	for (p1 = str; *p1; p1++)
	{
		if (*p1 == '/')
			*p1 = '_';
	}

	return str;
}


/*
 * This stuff doesnt conform to the protocol.
 * Thanks mirc for disregarding the protocol.
 */
#ifdef MIRC_BROKEN_DCC_RESUME

/*
 * Usage: /DCC RESUME <nick> [file] [-e passkey]
 */
static	void	dcc_getfile_resume (char *args)
{
	char		*user;
	char		*filename = NULL;
	DCC_list	*Client;
	char		*passwd = NULL;
	struct stat	sb;
	int		old_dp, old_dn, old_dc;

	if (!get_int_var(MIRC_BROKEN_DCC_RESUME_VAR))
	{
		say("DCC RESUME is turned off");
		return;
	}

	if (!(user = next_arg(args, &args)))
	{
		set_display_target(NULL, LOG_DCC);
		say("You must supply a nickname for DCC RESUME");
		reset_display_target();
		return;
	}

	if (args && *args)
	{
		/* Leeme lone, Yoshi. :P */
		if (args[0] != '-' || args[1] != 'e')
			filename = next_arg(args, &args);

		if (args && args[0] == '-' && args[1] == 'e')
		{
			next_arg(args, &args);
			passwd = next_arg(args, &args);
		}
	}

	/*
	 * This has to be done by hand, we cant use send_ctcp,
	 * because this violates the protocol, and send_ctcp checks
	 * for that.  Ugh.
	 */
	if (stat(filename, &sb) == -1)
	{
		/* File doesnt exist.  Sheesh. */
		set_display_target(NULL, LOG_DCC);
		say("DCC RESUME: Cannot use DCC RESUME if the file "
			"doesnt exist. [%s|%s]", filename, strerror(errno));
		reset_display_target();
		return;
	}


	if (!(Client = dcc_searchlist(filename, user, 
					DCC_FILEREAD, 0, NULL, 0)))
	{
		set_display_target(NULL, LOG_DCC);
		if (filename)
			say("No file (%s) offered in SEND mode by %s", 
					filename, user);
		else
			say("No file offered in SEND mode by %s", user);
		reset_display_target();

		return;
	}

	if ((Client->flags & DCC_ACTIVE) || (Client->flags & DCC_MY_OFFER))
	{
		set_display_target(NULL, LOG_DCC);
		say("A previous DCC GET:%s to %s exists", 
				filename ? filename : "<any>", user);
		reset_display_target();

		return;
	}

	if (passwd)
		Client->encrypt = m_strdup(passwd);
	Client->bytes_sent = 0L;
	Client->bytes_read = sb.st_size;

	malloc_strcpy(&Client->othername, 
			ltoa((long)ntohs(Client->remote_port)));

	if (x_debug & DEBUG_DCC_XMIT)
	{
		set_display_target(NULL, LOG_DCC);
		yell("SENDING DCC RESUME to [%s] [%s|%d|%ld]", 
			user, filename, ntohs(Client->remote_port), 
			(long)sb.st_size);
		reset_display_target();
	}

	/* Just in case we have to fool the protocol enforcement. */
	old_dp = doing_privmsg();
	old_dn = doing_notice();
	old_dc = doing_ctcp();

	set_doing_privmsg(0);
	set_doing_notice(0);
	set_doing_ctcp(0);

	send_ctcp(CTCP_PRIVMSG, user, CTCP_DCC, "RESUME %s %d %ld", 
		filename, ntohs(Client->remote_port), (long)sb.st_size);

	set_doing_privmsg(old_dp);
	set_doing_notice(old_dn);
	set_doing_ctcp(old_dc);

	/* Then we just sit back and wait for the reply. */
}

/*
 * When the peer demands DCC RESUME
 * We send out a DCC ACCEPT
 */
static void 	dcc_getfile_resume_demanded (char *user, char *filename, char *port, char *offset)
{
	DCC_list	*Client;
	int		old_dp, old_dn, old_dc;

	if (!get_int_var(MIRC_BROKEN_DCC_RESUME_VAR))
		return;

	if (x_debug & DEBUG_DCC_XMIT)
	{
		set_display_target(NULL, LOG_DCC);
		yell("GOT DCC RESUME REQUEST from [%s] [%s|%s|%s]", 
				user, filename, port, offset);
		reset_display_target();
	}

	if (!strcmp(filename, "file.ext"))
		filename = NULL;

	if (!(Client = dcc_searchlist(filename, user, 
					DCC_FILEOFFER, 0, port, 0)))
	{
		if (x_debug & DEBUG_DCC_XMIT)
		{
			set_display_target(NULL, LOG_DCC);
			yell("Resume request that doesnt exist.  Hmmm.");
			reset_display_target();
		}
		return;		/* Its a fake. */
	}

	if (!offset)
		return;		/* Its a fake */

	Client->bytes_sent = my_atol(offset);
	Client->bytes_read = 0L;

	/* Just in case we have to fool the protocol enforcement. */
	old_dp = doing_privmsg();
	old_dn = doing_notice();
	old_dc = doing_ctcp();

	set_doing_privmsg(0);
	set_doing_notice(0);
	set_doing_ctcp(0);

	send_ctcp(CTCP_PRIVMSG, user, CTCP_DCC, "ACCEPT %s %s %s",
		filename, port, offset);

	set_doing_privmsg(old_dp);
	set_doing_notice(old_dn);
	set_doing_ctcp(old_dc);

	/* Wait for them to open the connection */
}


/*
 * When we get the DCC ACCEPT
 * We start the connection
 */
static	void	dcc_getfile_resume_start (char *nick, char *filename, char *port, char *offset)
{
	DCC_list	*Client;
	char		*fullname;

	if (!get_int_var(MIRC_BROKEN_DCC_RESUME_VAR))
		return;

	if (x_debug & DEBUG_DCC_XMIT)
	{
		set_display_target(NULL, LOG_DCC);
		yell("GOT CONFIRMATION FOR DCC RESUME from [%s] [%s]", 
				nick, filename);
		reset_display_target();
	}

	if (!strcmp(filename, "file.ext"))
		filename = NULL;

	if (!(Client = dcc_searchlist(filename, nick, 
					DCC_FILEREAD, 0, port, 0)))
		return;		/* Its fake. */

	Client->flags |= DCC_TWOCLIENTS;
	if (!dcc_open(Client))
		return;

	if (!(fullname = expand_twiddle(Client->description)))
		malloc_strcpy(&fullname, Client->description);

	if (!(Client->file = open(fullname, O_WRONLY | O_APPEND, 0644)))
	{
		Client->flags |= DCC_DELETE;

		set_display_target(NULL, LOG_DCC);
		say("Unable to open %s: %s", 
			fullname, errno ? strerror(errno) : "<No Error>");
		reset_display_target();
	}

	new_free(&fullname);
}

#endif
