/* $EPIC:$ */
/*
 * notice.c: special stuff for parsing NOTICEs
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993-1999 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include "irc.h"
#include "ctcp.h"
#include "window.h"
#include "lastlog.h"
#include "flood.h"
#include "vars.h"
#include "ircaux.h"
#include "hook.h"
#include "ignore.h"
#include "server.h"
#include "output.h"
#include "names.h"
#include "parse.h"
#include "notify.h"
#include "commands.h"
#include "notice.h"

static	time_t 	convert_note_time_to_real_time (char *stuff);
static	int 	kill_message 		(const char *from, const char *line);
static void 	parse_local_server_notice (const char *from, char *to, const char *line);
static 	void 	parse_note 		(char *from, char *line);

/*
 * The main handler for those wacky NOTICE commands...
 */
void 	parse_notice (char *from, char **Args)
{
	int	type;
	char	*to;
	char	*high;
	char	*line;
	char	*target;
	int	no_flooding;

	PasteArgs(Args, 1);
	to = Args[0];
	line = Args[1];
	if (!to || !line)
		return;

	/*
	 * If 'to' is empty (present, but zero length), this is a seriously
	 * mal-formed NOTICE.  I think this test is here only to save a 
	 * coredump -- this shouldn't happen with any real irc server.
	 */
	if (!*to)
	{
		if (line && *line)
			put_it("[obsolete server notice] %s", line + 1);
		return;
	}

	/*
	 * Suppress the sending of PRIVMSGs or NOTICEs until this 
	 * value is reset.
	 */
	set_doing_notice(1);

	/*
	 * Note that NOTICEs from servers are not "server notices" unless
	 * the target is not a channel (ie, it is sent to us).  Any notice
	 * that is sent to a channel is a normal NOTICE, notwithstanding
	 * _who_ sent it.
	 */
	if (is_channel(to))
	{
		target = to;
		type = PUBLIC_NOTICE_LIST;
	}
	else
	{
		/*
		 * Now we see if the NOTICE is from our local server, and we
		 * call the server notice handler if it is.
		 */
		if (!from || !*from || 
			!strcmp(get_server_itsname(from_server), from))
		{
			parse_local_server_notice(from, to, line);
			set_doing_notice(0);
			return;
		}
		else
		{
			target = from;
			type = NOTICE_LIST;
		}
	}


	set_display_target(target, LOG_NOTICE);

	/*
	 * Check to see if we want to ignore the person, or highlight
	 * their message, or whatever.
	 */
	switch (check_ignore_channel(from, FromUserHost, to, IGNORE_NOTICES))
	{
		case IGNORED:
			goto notice_cleanup;
		case HIGHLIGHTED:
			high = highlight_char;
			break; /* oops! */
		default:
			high = empty_string;
	}

	/* Check for /notify's */
	notify_mark(from, 1, 0);

	/* Do normal /CTCP reply handling */
	line = do_notice_ctcp(from, to, line);
	if (!*line)
		goto notice_cleanup;

	/* Check for flooding */
	no_flooding = check_flooding(from, NOTICE_FLOOD, line);

	/*
	 * Check to see if the message was encrypted, and if it was, offer
	 * it to the user as such.
	 */
	if (sed == 1 && !do_hook(ENCRYPTED_NOTICE_LIST, "%s %s %s", 
						from, to, line))
	{
		sed = 0;
		goto notice_cleanup;
	}

	/* Do not parse the notice if we are being flooded */
	if (!no_flooding)
		goto notice_cleanup;

	/* Now offer the notice to the user and output if they dont take it. */
	if (type == NOTICE_LIST)
	{
		if (do_hook(type, "%s %s", from, line))
			put_it("%s-%s-%s %s", high, from, high, line);
	}
	else
	{
		if (do_hook(type, "%s %s %s", from, to, line))
			put_it("%s-%s:%s-%s %s", high, from, to, high, line);
	}

	if (beep_on_level & LOG_NOTICE)
		beep_em(1);

notice_cleanup:
	set_doing_notice(0);
	reset_display_target();
	return;
}


/*
 * This handles rfc1459 NOTICEs that have been sent from the server to
 * which we are connected -- NOTICEs from remote servers are treated as
 * normal NOTICEs from users.
 */
static void 	parse_local_server_notice (const char *from, char *to, const char *line)
{
	int		hooked = 0;
	const char *	r = from;

	if ((!r || !*r) && !(r = get_server_itsname(from_server)))
		r = get_server_name(from_server);

	/*
	 * OPERATOR_NOTICEs are a special kind of NOTICE from the server.
	 * The reason they are special is because the server has its own
	 * specific routine to spit these out -- to the server, these are
	 * a distinct, seperate message type; the server just uses NOTICE
	 * as its transport mechanism for this message type.  So it is useful
	 * for us to parse out this message and offer it to the user as it
	 * was intended.
	 */
	if (!strncmp(line, "*** Notice -- ", 13))
	{
		/*
		 * One of the OPERATOR_NOTICEs is the KILL notification,
		 * which many operators want to be able to handle easily.
		 */
		if (!strncmp(line + 14, "Received KILL message for ", 26))
			hooked = kill_message(r, line + 40);

		if (!hooked)
		{
			set_display_target(to, LOG_OPNOTE);
			if (!do_hook(OPER_NOTICE_LIST, "%s %s", r, line + 14))
				hooked = 1;
			reset_display_target();
		}

		return;
	}

	/*
	 * NOTEs are a special kind of NOTICEs that precede the days of
	 * the RFC.  They really are quite unlike anything else, but 
	 * it is useful to parse them as they still technically do exist
	 * in 2.8 servers, even though nobody uses them.
	 */
	if (!strncmp(line, "Note from ", 10) && line[10])
	{
		char *	note_from = NULL;
		char *	my_line;

		note_from = LOCAL_COPY(line + 10);
		if ((my_line = strchr(note_from, '!')))
		{
			*my_line++ = 0;
			FromUserHost = my_line;

			if ((my_line = strchr(FromUserHost, ' ')))
			{
				*my_line++ = 0;
				parse_note(note_from, my_line);
			}
			FromUserHost = empty_string;
		}
		return;
	}

	/*
	 * So these are just normal NOTICEs from our server.
	 */
	set_display_target(from, LOG_SNOTE);

	/*
	 * All server notices need to have a "header" included, but some of
	 * the server notices have their own headers, and others do not.
	 * So we peek and see if this one has a header and if it does we use
	 * it and if it doesn't, we supply our own.
	 */
	if (*line == '*' || *line == '#')
	{
		if (!hooked && do_hook(SERVER_NOTICE_LIST, "%s %s", r, line))
			put_it("%s", line);
	}
	else if (!hooked && do_hook(SERVER_NOTICE_LIST, "%s *** %s", r, line))
		say("%s", line);

	reset_display_target();
}


static int 	kill_message (const char *from, const char *cline)
{
	char *	poor_sap;
	char *	bastard;
	char *	path_to_bastard;
	char *	reason;
	char *	line;

	line = LOCAL_COPY(cline);
	poor_sap = next_arg(line, &line);

	/* Broken Dalnet kill doesnt append the period */
	if (!end_strcmp(poor_sap, ".", 1))
		chop(poor_sap, 1);

	/* Broken dalnet kill doesnt use "From", but "from" */
	if (my_strnicmp(line, "From ", 5))
	{
		yell("Attempted to parse an ill-formed KILL request [%s %s]",
			poor_sap, line);
		return 0;
	}
	line += 5;
	bastard = next_arg(line, &line);

	/* Fend off future broken kills */
	if (my_strnicmp(line, "Path: ", 6))
	{
		yell("Attempted to parse an ill-formed KILL request [%s %s %s]",
			poor_sap, bastard, line);
		return 0;
	}
	line += 6;
	path_to_bastard = next_arg(line, &line);
	reason = line;

	return !do_hook(KILL_LIST, "%s %s %s %s %s", from, poor_sap, bastard,
					path_to_bastard, reason);
}



/*
 * The client apparantly never was adapted to handle the new NOTE syntax.
 * So i had to kludge this up to work with it.  Currently, NOTEs are sent
 * something like this:
 *
 *	NOTICE yournick :Note from nick!user@host /xxxd:xxh:xxm/ [N] message
 *
 * and parse() calls parse_notice(), who notices that there is no pefix
 * and passes it off to parse_local_server_notice(), who checks to see if it is
 * a note (it is), blows away the "Note from" part, and re-sets the "from"
 * and "FromUserHost" parts with the nick!user@host part and passes us the
 * buck with 'line' pointing at the time description (the /.../ part)
 */
static void 	parse_note (char *from, char *line)
{
	char	*date,
		*flags,
		*high;
	time_t	when;

	switch (check_ignore(from, FromUserHost, IGNORE_NOTES))
	{
		case IGNORED:
			return;
		case HIGHLIGHTED:
			high = highlight_char;
			break;
		default:
			high = empty_string;
	}

	if (!check_flooding(from, NOTE_FLOOD, line))
		return;

	/* 
	 * at this point, line looks like:
	 * "/xxxd:xxh:xxm/ [FLAGS] message goes here"
	 */
	date = next_arg(line, &line);
	flags = next_arg(line, &line);
	when = convert_note_time_to_real_time(date);

	set_display_target(from, LOG_NOTES);
	if (do_hook(NOTE_LIST, "%s %lu %s", from, when, line))
	{
		if (time(NULL) - when > 60)	/* not just sent */
			put_it("%s[%s]%s %s (%s)", 
				high, from, high, line, my_ctime(when));
		else
			put_it("%s[%s]%s %s", high, from, high, line);
	}
	reset_display_target();
	if (beep_on_level & LOG_NOTES)
		beep_em(1);

	return;
}

static time_t 	convert_note_time_to_real_time (char *stuff)
{
	time_t 	days = 0, 
		hours = 0, 
		minutes = 0;

	stuff++;			      /* first character is a '/' */
	days = strtoul(stuff, &stuff, 10);    /* get the number of days */
	stuff++;			      /* skip over the 'd' */
	stuff++;			      /* skip over the ':' */
	hours = strtoul(stuff, &stuff, 10);   /* get the number of hours */
	stuff++;			      /* skip over the 'h' */
	stuff++;			      /* skip over the ':' */
	minutes = strtoul(stuff, &stuff, 10); /* get the number of minutes */
	stuff++;			      /* skip over the 'm' */
	stuff++;			      /* skip over the '/' */
	if (*stuff)
		yell("cntto: bad format");

	hours += days * 24;
	minutes += hours * 60;
	return (time(NULL) - minutes * 60);
}


/*
 * got_initial_version_28: this is called when ircii gets the serial
 * number 004 reply.  We do this becuase the 004 numeric gives us the
 * server name and version in an easy to use fashion, and doesnt
 * rely on the syntax or construction of the 002 numeric.
 *
 * Hacked as neccesary by jfn, May 1995
 * XXX This probably doesn't belong here.
 */
void 	got_initial_version_28 (char **ArgList)
{
	char *server, *version, *umodes;

	reset_display_target();

	server = ArgList[0];
	version = ArgList[1];
	umodes = ArgList[2];

	if (!strncmp(version, "2.8", 3))
	{
		if (strstr(version, "mu") || strstr(version, "me"))
			set_server_version(from_server, Server_u2_8);
		else
			set_server_version(from_server, Server2_8);
	}
	else if (!strncmp(version, "2.9", 3))
		set_server_version(from_server, Server2_9);
	else if (!strncmp(version, "u2.9", 4))
		set_server_version(from_server, Server_u2_9);
	else if (!strncmp(version, "u2.10", 4))
		set_server_version(from_server, Server_u2_10);
	else if (!strncmp(version, "u3.0", 4))
		set_server_version(from_server, Server_u3_0);
	else
		set_server_version(from_server, Server2_8);

	set_server_version_string(from_server, version);
	set_server_itsname(from_server, server);
	channel_server_reconnect(from_server);
	reinstate_user_modes(from_server);

	if (never_connected)
	{
		never_connected = 0;
		if (!ircrc_loaded)
			load_ircrc();

		if (default_channel)
		{
			char *blah = alloca(strlen(default_channel) + 4);
			sprintf(blah, "\"%s\"", default_channel);
			window_channel(current_window, &blah);
			new_free(&default_channel);
		}
	}
	else if (get_server_away(from_server))
		set_server_away(from_server, get_server_away(from_server));

	update_all_status();
	do_hook(CONNECT_LIST, "%s %d %s", 
		get_server_name(from_server),
		get_server_port(from_server), 
		get_server_itsname(from_server));

	return;
}


/*
 * XXX - I suppose this doesn't belong here, but where does it belong?
 */
void	load_ircrc (void)
{
	char buffer[7];
	strcpy(buffer, "global");

	loading_global = 1;
	load("LOAD", buffer, empty_string);
	loading_global = 0;

	/* read the startup file */
	if (access(epicrc_file, R_OK) == 0 && !quick_startup)
	{
		load("LOAD", epicrc_file, empty_string);
		startup_file = epicrc_file;
	}
	else if (access(ircrc_file, R_OK) == 0 && !quick_startup)
	{
		load("LOAD", ircrc_file, empty_string);
		startup_file = ircrc_file;
	}

	ircrc_loaded = 1;
}

