/* $EPIC:$ */
/*
 * screen.c -- device independant output handlers.
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993-2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include "irc.h"
#include "ircaux.h"
#include "screen.h"
#include "window.h"
#include "output.h"
#include "vars.h"
#include "server.h"
#include "input.h"
#include "status.h"
#include "commands.h"
#include "newio2.h"
#include "network.h"
#include "term.h"
#include "exec.h"

#define CURRENT_WSERV_VERSION	4

/*
 * When all else fails, this is the screen that is attached to the controlling
 * terminal, and we know *that* screen will always be there.
 */
	Screen	*main_screen;

/*
 * This is the screen in which we last handled an input event.  This takes
 * care of the input duties that "current_screen" used to handle.
 */
	Screen	*last_input_screen;

/*
 * This is used to set the default output device for tputs_x().  This takes
 * care of the output duties that "current_screen" used to handle.  Since
 * the input screen and output screen are independant, it isnt impossible 
 * for a command in one screen to cause output in another.
 */
	Screen	*output_screen;

/*
 * The list of all the screens we're handling.  Under most cases, there's 
 * only one screen on the list, "main_screen".
 */
	Screen	*screen_list = NULL;

static	int	wserv_callback	    (int, int, char *);
static	int	wserv_error	    (int);
static	int	screen_callback	    (int, int, char *);
static	int	screen_error	    (int);

Discipline	wserv_discipline = {
	wserv_callback,
	NULL,
	wserv_error,
	wserv_error,
};

Discipline	screen_discipline = {
	screen_callback,
	NULL,
	screen_error,
	screen_error
};


/* * * * * * * * * * * * * * SCREEN MANAGEMENT * * * * * * * * * * * * */
/*
 * create_new_screen creates a new screen structure. with the help of
 * this structure we maintain ircII windows that cross screen window
 * boundaries.
 */
Screen *	create_new_screen (void)
{
	Screen	*new_s = NULL, *list;
	static	int	refnumber = 0;

	for (list = screen_list; list; list = list->next)
	{
		if (!list->alive)
		{
			new_s = list;
			break;
		}

		if (!list->next)
			break;		/* XXXX */
	}

	if (!new_s)
	{
		new_s = (Screen *)new_malloc(sizeof(Screen));
		new_s->screennum = ++refnumber;
		new_s->next = NULL;
		if (list)
			list->next = new_s;
		else
			screen_list = new_s;
	}

	new_s->last_window_refnum = 1;
	new_s->window_list = NULL;
	new_s->window_list_end = NULL;
	new_s->cursor_window = NULL;
	new_s->current_window = NULL;
	new_s->visible_windows = 0;
	new_s->window_stack = NULL;
	new_s->meta_hit = 0;
	new_s->quote_hit = 0;
	new_s->fdin = 0;
	new_s->fpin = stdin;
	new_s->fdout = 1;
	new_s->fpout = stdout;
	new_s->control = -1;
	new_s->wserv_version = 0;
	new_s->alive = 1;
	new_s->promptlist = NULL;
	new_s->tty_name = NULL;
	new_s->li = current_term->TI_lines;
	new_s->co = current_term->TI_cols;
	new_s->old_li = 0; 
	new_s->old_co = 0;

	new_s->buffer_pos = new_s->buffer_min_pos = 0;
	new_s->input_buffer[0] = '\0';
	new_s->input_cursor = 0;
	new_s->input_visible = 0;
	new_s->input_start_zone = 0;
	new_s->input_end_zone = 0;
	new_s->input_prompt = NULL;
	new_s->input_prompt_len = 0;
	new_s->input_prompt_malloc = 0;
	new_s->input_line = 23;

	last_input_screen = new_s;

	if (!main_screen)
	{
		main_screen = new_s;
		no_buffering_readonly(0, &screen_discipline);
	}

	init_input();
	return new_s;
}


#define ST_NOTHING	-1
#define ST_SCREEN	0
#define ST_XTERM	1

Window *	create_additional_screen (void)
{
        Window  	*win;
        Screen  	*oldscreen, *new_s;
        char    	*displayvar,
                	*termvar;
        int     	screen_type = ST_NOTHING;
        ISA		new_socket;
	int		new_cmd;
        int     	s;
	fd_set		fd_read;
	Timeval		timeout;
	pid_t		child;
	unsigned short 	port;
	int		new_sock_size;
	char *		wserv_path;

	if (!use_input)
		return NULL;

	if (!(wserv_path = get_string_var(WSERV_PATH_VAR)))
	{
		say("You need to /SET WSERV_PATH before using /WINDOW CREATE");
		return NULL;
	}

	/* Check for X first. */
	if ((displayvar = getenv("DISPLAY")))
	{
		if (!(termvar = getenv("TERM")))
			say("I don't know how to create new windows "
				"for this terminal");
		else
			screen_type = ST_XTERM;
	}

	/*
	 * Environment variable STY has to be set for screen to work..  so it is
	 * the best way to check screen..  regardless of what TERM is, the 
	 * execpl() for screen won't just open a new window if STY isn't set,
	 * it will open a new screen process, and run the wserv in its first
	 * window, not what we want...  -phone
	 */
	if (screen_type == ST_NOTHING && getenv("STY"))
		screen_type = ST_SCREEN;


	if (screen_type == ST_NOTHING)
	{
		say("I don't know how to create new windows for this terminal");
		return NULL;
	}

        say("Opening new %s...",
                screen_type == ST_XTERM ?  "window" :
                screen_type == ST_SCREEN ? "screen" :
                                           "wound" );

	port = 0;
	new_cmd = connect_by_number(NULL, &port, SERVICE_SERVER, 
						 PROTOCOL_TCP, NULL);
	if (new_cmd < 0)
	{
		yell("Couldnt establish server side -- error [%d]", new_cmd);
		return NULL;
	}

	oldscreen = current_window->screen;
	new_s = create_new_screen();

	switch ((child = fork()))
	{
		case -1:
		{
			kill_screen(new_s);
			say("I couldnt fire up a new wserv process");
			break;
		}

		case 0:
		{
			char *opts;
			char *xterm;
			char *args[64], **args_ptr = args;
			char geom[32];
			int i;

			setuid(getuid());
			setgid(getgid());
			setsid();

			/*
			 * Make sure that no inhereted file descriptors
			 * are left over past the exec.  xterm will reopen
			 * any fd's that it is interested in.
			 * (Start at three sb kanan).
			 */
			for (i = 3; i < 256; i++)
				close(i);

			/*
			 * Try to restore some sanity to the signal
			 * handlers, since theyre not really appropriate here
			 */
			my_signal(SIGINT,  SIG_IGN);
			my_signal(SIGSEGV, SIG_DFL);
			my_signal(SIGBUS,  SIG_DFL);
			my_signal(SIGABRT, SIG_DFL);

			if (screen_type == ST_SCREEN)
			{
			    opts = m_strdup(get_string_var(SCREEN_OPTIONS_VAR));
			    *args_ptr++ = "screen";
			    while (opts && *opts)
				*args_ptr++ = new_next_arg(opts, &opts);
			}
			else if (screen_type == ST_XTERM)
			{
			    snprintf(geom, 31, "%dx%d", 
				oldscreen->co + 1, oldscreen->li);
			    opts = m_strdup(get_string_var(XTERM_OPTIONS_VAR));
			    if (!(xterm = getenv("XTERM")))
				if (!(xterm = get_string_var(XTERM_VAR)))
				    xterm = "xterm";

			    *args_ptr++ = xterm;
			    *args_ptr++ = "-geometry";
			    *args_ptr++ = geom;
			    while (opts && *opts)
				*args_ptr++ = new_next_arg(opts, &opts);
			    *args_ptr++ = "-e";
			}

			*args_ptr++ = wserv_path;
			*args_ptr++ = "localhost";
			*args_ptr++ = m_strdup(ltoa((long)port));
			*args_ptr++ = NULL;

			s = execvp(args[0], args);
			_exit(0);
		}
	}

	/* All the rest of this is the parent.... */
	new_sock_size = sizeof(new_socket);
	FD_ZERO(&fd_read);
	FD_SET(new_cmd, &fd_read);
	timeout.tv_sec  = 10;
	timeout.tv_usec =  0;

	/* 
	 * This infinite loop sb kanan to allow us to trap transitory
	 * error signals
	 */
	for (;;)

	/*
	 * At this point, doing a say() or a yell() or anything else that
	 * would output to the screen will cause a refresh of the status bar
	 * and input_line.  new_s->current_window is NULL after the above 
	 * line and so any attempt to reference $C or $T will be to a NULL
	 * pointer, which will cause a crash.  For various resaons, we can't
	 * fire up a new window this early, so its just easier to make sure
	 * we don't output anything until kill_screen() or new_window() is
	 * called first.  You have been warned!
	 */
	switch (select(new_cmd + 1, &fd_read, NULL, NULL, &timeout))
	{
	    case -1:
	    {
		if ((errno == EINTR) || (errno == EAGAIN))
			continue;
		/* FALLTHROUGH */
	    }
	    case 0:
	    {
		int 	old_errno = errno;
		int 	errnod = get_child_exit(child);

		close(new_cmd);
		kill_screen(new_s);
		kill(child, SIGKILL);

		if (new_s->fdin != 0)
		{
			say("The wserv only connected once -- it's probably "
			    "an old, incompatable version.");
		}

                yell("child %s with %d", (errnod < 1) ? "signaled" : "exited",
                                         (errnod < 1) ? -errnod : errnod);
		yell("Errno is %d", old_errno);
		return NULL;
	    }
	    default:
	    {
		if (new_s->fdin == 0)
		{
			new_s->fdin = accept(new_cmd, (SA *)&new_socket, 
						&new_sock_size);

			if ((new_s->fdout = new_s->fdin) < 0)
			{
				close(new_cmd);
				kill_screen(new_s);
				yell("Couldn't establish data connection "
					"to new screen");
				return NULL;
			}

			no_buffering_readonly(new_s->fdin, &screen_discipline);
			new_s->fpin = new_s->fpout = fdopen(new_s->fdin, "r+");
			continue;
		}
		else
		{
			new_s->control = accept(new_cmd, (SA *)&new_socket, 
						&new_sock_size);
			close(new_cmd);
			if (new_s->control < 0)
			{
				kill_screen(new_s);
				yell("Couldn't establish control connection "
					"to new screen");
				return NULL;
			}

			line_buffering(new_s->control, &wserv_discipline);

			if (!(win = new_window(new_s)))
				panic("WINDOW is NULL and it shouldnt be!");
			return win;
		}
	    }
	}
	return NULL;
}

/* Old screens never die. They just fade away. */
void 	kill_screen (Screen *screen)
{
	Window	*window;

	if (!screen)
	{
		say("You may not kill the hidden screen.");
		return;
	}
	if (main_screen == screen)
	{
		say("You may not kill the main screen");
		return;
	}
	if (screen->fdin)
	{
		screen->fdin = new_close(screen->fdin);
		screen->control = new_close(screen->control);
		screen->fdout = new_close(screen->fdout);
	}
	while ((window = screen->window_list))
	{
		screen->window_list = window->next;
		add_to_invisible_list(window);
	}

	/* Dont fool around. */
	if (last_input_screen == screen)
		last_input_screen = main_screen;

	/* Take out some of the garbage left around */
	screen->current_window = NULL;
	screen->window_list = NULL;
	screen->window_list_end = NULL;
	screen->cursor_window = NULL;
	screen->last_window_refnum = -1;
	screen->visible_windows = 0;
	screen->window_stack = NULL;
	screen->fpin = NULL;
	screen->fpout = NULL;
	screen->fdin = -1;
	screen->fdout = -1;
	new_free(&screen->input_prompt);

	/* Don't fool around. */
	if (last_input_screen == screen)
		last_input_screen = main_screen;

	screen->alive = 0;
	make_window_current(NULL);
	say("The screen is now dead.");
}


/* * * * * * * * * * * * * USER INPUT HANDLER * * * * * * * * * * * */
static int	screen_error	(int des)
{
	Screen	*screen;
	int	server;

	for (screen = screen_list; screen; screen = screen->next)
		if (screen->fdin == des)
			break;

	server = from_server;
	last_input_screen = screen;
	output_screen = screen;
	make_window_current(screen->current_window);
	from_server = current_window->server;

	if (dumb_mode)
	{
		say("IRCII exiting on EOF from stdin");
		irc_exit(1, "EPIC - EOF from stdin");
	}

	/*
	 * if the current screen isn't the main screen,
	 * then the socket to the current screen must have
	 * closed, so we call kill_screen() to handle 
	 * this - phone, jan 1993.
	 * but not when we arent running windows - Fizzy, may 1993
	 * if it is the main screen we got an EOF on, we exit..
	 * closed tty -> chew cpu -> bad .. -phone, july 1993.
	 */
	else if (screen != main_screen)
		kill_screen(screen);

	/*
	 * If we got an error at this point, then something just totaly
	 * died on us.  This is usually caused by the ctty being
	 * revoke(2)d on us.  4.4BSD guarantees that a revoke()d ctty 
	 * will read an EOF, while i believe linux fails with EBADF.  
	 * In either case, a read failure on the main screen is totaly fatal.
	 */
	else
		irc_exit(1, "Hey!  Where'd my controlling terminal go?");

	from_server = server;
	return 0;
}

static int	screen_callback	(int des, int len, char *buffer)
{
	Screen	*screen;
	int 	server;
	int	i;

	for (screen = screen_list; screen; screen = screen->next)
		if (screen->fdin == des)
			break;

	if (screen != main_screen && screen->wserv_version == 0)
	{
		yell("The WSERV used to create this new screen is too old.");
		kill_screen(screen);
		return 1;
	}

	/*
	 * This section of code handles all in put from 
	 * the terminal(s) connected to ircII.  Perhaps the 
	 * idle time *shouldn't* be reset unless its not a 
	 * screen-fd that was closed..
	 */
	time(&idle_time);
	if (cpu_saver)
	{
		cpu_saver = 0;
		update_all_status();
	}

	server = from_server;
	last_input_screen = screen;
	output_screen = screen;
	make_window_current(screen->current_window);
	from_server = current_window->server;

	if (dumb_mode)
	{
		buffer[len] = 0;
		if (get_int_var(INPUT_ALIASES_VAR))
			parse_line(NULL, buffer, empty_string, 1, 0);
		else
			parse_line(NULL, buffer, NULL, 1, 0);

		return 0;
	}

	for (i = 0; i < len; i++)
		edit_char(buffer[i]);

	from_server = server;
	return 0;
}

static int	wserv_error (int des)
{
	Screen	*screen;

	for (screen = screen_list; screen; screen = screen->next)
		if (screen->control == des)
			break;

	kill_screen(screen);
	yell("Error from remote screen.");
	return 0;
}

static int	wserv_callback (int des, int len, char *buffer)
{
	Screen	*screen;

	for (screen = screen_list; screen; screen = screen->next)
		if (screen->control == des)
			break;

	buffer[len] = 0;
	if (!strncmp(buffer, "tty=", 4))
		malloc_strcpy(&screen->tty_name, buffer + 4);
	else if (!strncmp(buffer, "geom=", 5))
	{
		char *ptr;
		if ((ptr = strchr(buffer, ' ')))
			*ptr++ = 0;
		screen->li = atoi(buffer + 5);
		screen->co = atoi(ptr);
		refresh_a_screen(screen);
	}
	else if (!strncmp(buffer, "version=", 8))
	{
		int	version;
		version = atoi(buffer + 8);
		if (version != CURRENT_WSERV_VERSION)
		{
		    yell("WSERV version %d is incompatable with this binary", 
				version);
		    kill_screen(screen);
		}
		screen->wserv_version = version;
	}
	return 0;
}

/* * * * * * * * * INPUT PROMPTS * * * * * * * * * * */
/* 
 * add_wait_prompt:  Given a prompt string, a function to call when
 * the prompt is entered.. some other data to pass to the function,
 * and the type of prompt..  either for a line, or a key, we add 
 * this to the prompt_list for the current screen..  and set the
 * input prompt accordingly.
 *
 * XXXX - maybe this belongs in input.c? =)
 */
void 	add_wait_prompt (char *prompt, void (*func)(), char *data, int type, int echo)
{
	WaitPrompt **AddLoc,
		   *New;

	New = (WaitPrompt *) new_malloc(sizeof(WaitPrompt));
	New->prompt = m_strdup(prompt);
	New->data = m_strdup(data);
	New->type = type;
	New->echo = echo;
	New->func = func;
	New->next = NULL;
	for (AddLoc = &current_window->screen->promptlist; *AddLoc;
			AddLoc = &(*AddLoc)->next);
	*AddLoc = New;
	if (AddLoc == &current_window->screen->promptlist)
		change_input_prompt(1);
}

