/* $EPIC:$ */
/*
 * window-front.c -- window to character-mode terminal output interface
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993-2000 Jeremy Nelson and others ("EPIC Software Labs").
 * Copyright  1998 J. Kean Johnston, used with permission
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#define __need_putchar_x__
#define __need_term_flush__
#include "irc.h"
#include "screen.h"
#include "window.h"
#include "output.h"
#include "vars.h"
#include "server.h"
#include "term.h"
#include "ircaux.h"
#include "input.h"
#include "hook.h"
#include "status.h"
#include "commands.h"
#include "alias.h"
#include "window-front.h"

/*
 * Ugh.  Dont ask.
 */
	int	normalize_never_xlate = 0;
	int	normalize_permit_all_attributes = 0;

/*
 * This file includes major work contributed by FireClown, and I am indebted
 * to him for the work he has graciously donated to the project.  The major
 * highlights of his work include:
 *
 * -- ^C codes have been changed to mIRC-order.  This is the order that
 *    BitchX uses as well, so those scripts that use ^CXX should work without
 *    changes now between epic and bitchx.
 * -- The old "ansi-order" ^C codes have been preserved, but in a different
 *    way.  If you do ^C30 through ^C37, you will set the foreground color
 *    (directly corresponding to the ansi codes for 30-37), and if you do 
 *    ^C40 through ^C47, you will set the background.  ^C50 through ^C57
 *    are reserved for bold-foreground, and blink-background.
 * -- $cparse() still outputs the "right" colors, so if you use $cparse(),
 *    then these changes wont affect you (much).
 * -- Colors and ansi codes are either graciously handled, or completely
 *    filtered out.  Anything that cannot be handled is removed, so there
 *    is no risk of dangerous codes making their way to your output.  This
 *    is accomplished by a low-grade ansi emulator that folds raw output 
 *    into an intermediate form which is used by the display routines.
 *
 * To a certain extent, the original code from  FireClown was not yet complete,
 * and it was evident that the code was in anticipation of some additional
 * future work.  We have completed much of that work, and we are very much
 * indebted to him for getting the ball rolling and supplying us with ideas. =)
 */


/* * * * * * * * * * * * * OUTPUT CHAIN * * * * * * * * * * * * * * * * * * *
 * To put a message to the "default" window, you must first call
 *	set_display_target(nick/channel, lastlog_level)
 * Then you may call 
 *	say(), output(), yell(), put_it(), put_echo(), etc.
 * When you are done, make sure to
 *	reset_display_target()
 *
 * To put a message to a specific, known window (you need it's refnum)
 * then you may just call directly:
 *	display_to(winref, ...)
 *
 * To put a series of messages to a specific, known window, (need it's refnum)
 * You must first call:
 *	message_to(winref)
 * Then you may call
 *	say(), ouitput(), yell(), put_it(), put_echo(), etc.
 * When you are done, make sure to
 *	message_to(-1);
 *
 * The 'display' (or 'display_to') functions are the main entry point for
 * all logical output from epic.  These functions then figure out what
 * window the output will go to and invoke its 'add' function.  From there,
 * whatever happens is implementation defined.
 *
 * This file implements the middle part of the "ircII window", everything
 * from the 'add' function to the low level terminal stuff.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static int 	rite 		    (Window *window, const unsigned char *str);
static void 	scroll_window 	    (Window *window);

/*
 * "Attributes" were an invention for epic4pre3, and the general idea was
 * to expunge from the output chain all of those nasty logical toggle settings
 * which never really did work correctly.  Rather than have half a dozen
 * functions all keep state about whether reverse/bold/underline/whatever is
 * on, or off, or what to do when it sees a toggle, we instead have one 
 * function (normalize_string) which walks the string *once* and outputs a
 * completely normalized output string.  The end result of this change is
 * that what were formally "toggle" attributes now are "always-on" attributes,
 * and to turn off an attribute, you need to do an ALL_OFF (^O) and then
 * turn on whatever attributes are left.  This is *significantly* easier
 * to parse, and yeilds much better results, at the expense of a few extra
 * bytes.
 *
 * Now on to the nitty gritty.  Every character has a fudamental set of
 * attributes that apply to it.  Each character has, by default, the same
 * set of fundamental attributes as the character before it.  In any case
 * where this is NOT true, an "attribute marker" is put into the normalized
 * output to indicate what the new fundamental attributes are.  These new
 * attributes continue to be used until another attribute marker is found.
 *
 * The "Attribute" structure is an internal structure that represents all
 * of the supported fundamental attributes.  This is the prefered method
 * for keeping state of the attributes of a line.  You can convert this
 * structure into an "attribute marker" by passing the string and an 
 * Attribute struct to 'display_attributes'.  The result is 5 bytes of
 * output, each byte has the high bit set (so str*() still work).  You can
 * also convert an Attribute struct to standard ircII attribute characters
 * by calling 'logical_attributes'.  The result will be an ALL_OFF (^O) 
 * followed by all of the attributes that are ON in the struct.  Finally,
 * you can suppress all attribute changes by calling ignore_attribute().
 * These functions are used by normalize_string() to for their appropriate
 * uses.
 *
 * You can read an attribute marker from a string and convert it back to
 * an Attribute struct by calling the read_attributes() function.  You can
 * actually perform the physical output operations neccesary to switch to
 * the values in an Attribute struct by calling term_attribute().  These
 * are used by various output routines for whatever reason.
 */
struct 	attributes {
	int	reverse;
	int	bold;
	int	blink;
	int	underline;
	int	altchar;
	int	color_fg;
	int	color_bg;
	int	fg_color;
	int	bg_color;
};
typedef struct attributes Attribute;

const char *all_off (void)
{
#ifdef NO_CHEATING
	Attribute 	a;
	static	char	retval[6];

	a->reverse = a->bold = a->blink = a->underline = a->altchar = 0;
	a->color_fg = a->fg_color = a->color_bg = a->bg_color = 0;
	display_attributes(retval, &a);
	return retval;
#else
	static	char	retval[6];
	retval[0] = '\006';
	retval[1] = retval[2] = retval[3] = retval[4] = 0x80;
	retval[5] = 0;
	return retval;
#endif
}

/* Put into 'output', an attribute marker corresponding to 'a' */
static size_t	display_attributes (u_char *output, Attribute *a)
{
	u_char	val1 = 0x80;
	u_char	val2 = 0x80;
	u_char	val3 = 0x80;
	u_char	val4 = 0x80;

	if (a->reverse)		val1 |= 0x01;
	if (a->bold)		val1 |= 0x02;
	if (a->blink)		val1 |= 0x04;
	if (a->underline)	val1 |= 0x08;
	if (a->altchar)		val1 |= 0x10;

	if (a->color_fg) {	val2 |= 0x01; val3 |= a->fg_color; }
	if (a->color_bg) {	val2 |= 0x02; val4 |= a->bg_color; }

	output[0] = '\006';
	output[1] = val1;
	output[2] = val2;
	output[3] = val3;
	output[4] = val4;
	output[5] = 0;
	return 5;
}
 
/* Put into 'output', logical characters so end result is 'a' */
static size_t	logic_attributes (u_char *output, Attribute *a)
{
	char	*str = output;
	size_t	count = 0;

	*str++ = ALL_OFF, count++;
	/* Colors need to be set first, always */
	if (a->color_fg)
	{
		*str++ = '\003', count++;
		*str++ = '3', count++;
		*str++ = '0' + a->fg_color, count++;
	}
	if (a->color_bg)
	{
		if (!a->color_fg)
			*str++ = '\003', count++;
		*str++ = ',', count++;
		*str++ = '4', count++;
		*str++ = '0' + a->bg_color, count++;
	}
	if (a->bold)
		*str++ = BOLD_TOG, count++;
	if (a->blink)
		*str++ = BLINK_TOG, count++;
	if (a->reverse)
		*str++ = REV_TOG, count++;
	if (a->underline)
		*str++ = UND_TOG, count++;
	if (a->altchar)
		*str++ = ALT_TOG, count++;
	return count;
}

/* Suppress any attribute changes in the output */
static size_t	ignore_attributes (u_char *output, Attribute *a)
{
	return 0;
}

/* Read an attribute marker from 'input', put results in 'a'. */
static int	read_attributes (const u_char *input, Attribute *a)
{
	if (!input)
		return -1;
	if (*input != '\006')
		return -1;
	if (!input[0] || !input[1] || !input[2] || !input[3] || !input[4])
		return -1;

	a->reverse = a->bold = a->blink = a->underline = a->altchar = 0;
	a->color_fg = a->fg_color = a->color_bg = a->bg_color = 0;

	input++;
	if (*input & 0x01)	a->reverse = 1;
	if (*input & 0x02)	a->bold = 1;
	if (*input & 0x04)	a->blink = 1;
	if (*input & 0x08)	a->underline = 1;
	if (*input & 0x10)	a->altchar = 1;

	input++;
	if (*input & 0x01) {	
		a->color_fg = 1; 
		a->fg_color = input[1] & 0x7F; 
	}
	if (*input & 0x02) {	
		a->color_bg = 1; 
		a->bg_color = input[2] & 0x7F; 
	}

	return 0;
}

/* Invoke all of the neccesary functions so output attributes reflect 'a'. */
static void	term_attribute (Attribute *a)
{
	term_all_off();
	if (a->reverse)		term_standout_on();
	if (a->bold)		term_bold_on();
	if (a->blink)		term_blink_on();
	if (a->underline)	term_underline_on();
	if (a->altchar)		term_altcharset_on();

	if (a->color_fg) {	if (a->fg_color > 7) abort(); 
				else term_set_foreground(a->fg_color); }
	if (a->color_bg) {	if (a->bg_color > 7) abort();
				else term_set_background(a->bg_color); }
}

/************************************************************************/
/*
 * The main injection point for output to ircII type windows
 */
void	general_win_add (Window *window, const unsigned char *str)
{
	char *		pend;
	int		must_free = 0;
	char *		strval;

	/* Check to see if redirect must occur */
	if (window->server >= 0 && get_server_redirect(window->server))
		if (redirect_text(window->server, 
			        get_server_redirect(window->server),
				str, NULL, 0))
			return;

	/* User is suppressing this line of output. */
	if (!do_hook(WINDOW_LIST, "%u %s", window->refnum, str))
		return;

	/* Perform any rewrite rules as neccesary */
	if ((pend = get_string_var(OUTPUT_REWRITE_VAR)))
	{
		char *	prepend_exp;
		char	argstuff[10240];
		int	args_flag;

		/* First, create the $* list for the expando */
		snprintf(argstuff, 10240, "%u %s",
			window->refnum, str);

		/* Now expand the expando with the above $* */
		prepend_exp = expand_alias(pend, argstuff,
					   &args_flag, NULL);

		str = prepend_exp;
		must_free = 1;
	}

	/* Normalize the line of output */
	strval = normalize_string(str, 0);
	if (must_free)
		new_free(&str);

	/* And pass it off to the window */
	window->display(window, strval);
	new_free(&strval);

	/*
	 * This used to be in rite(), but since rite() is a general
	 * purpose function, and this stuff really is only intended
	 * to hook on "new" output, it really makes more sense to do
	 * this here.  This also avoids the terrible problem of 
	 * recursive calls to split_up_line, which are bad.
	 */
	if (!window->visible)
	{
		/*
		 * This is for archon -- he wanted a way to have 
		 * a hidden window always beep, even if BEEP is off.
		 */
		if (window->beep_always && strchr(str, '\007'))
		{
			Winref	old;

			term_beep();
			old = message_to(0);
			say("Beep in window %d", window->refnum);
			message_to(old);
		}

		/*
		 * Tell some visible window about our problems 
		 * if the user wants us to.
		 */
		if (!window->miscflags & WINDOW_NOTIFIED &&
			(output_level() & window->notify_level))
		{
			window->miscflags |= WINDOW_NOTIFIED;

			if (window->miscflags & WINDOW_NOTIFY &&
				output_level() & window->notify_level)
			{
				Winref	old;

				old = message_to(0);
				say("Activity in window %d", window->refnum);
				message_to(old);
				update_all_status();
			}
		}
	}
}

/*
 * Main display routine for ircII windows
 */
void	irciiwin_display (Window *window, const unsigned char *str)
{
	u_char **	lines;
	int		cols;
	int		numl;

	/* Add it to our lastlog */
	window->lastlog(window, str);

	/* Break it up and add it to our display */
	if (window->screen)
		cols = window->screen->co;
	else
		cols = window->columns;

	/* Suppress status updates while we're outputting. */
	for (lines = prepare_display(str, cols, &numl, 0); *lines; lines++)
	{
		if (add_to_display_list(window, *lines))
			rite(window, *lines);
	}

	term_flush();
	cursor_to_input();
}

/* * * * * * * * * * * * * COLOR SUPPORT * * * * * * * * * * * * * * * * */
/*
 * This parses out a ^C control sequence.  Note that it is not acceptable
 * to simply slurp up all digits after a ^C sequence (either by calling
 * strtol(), or while (isdigit())), because people put ^C sequences right
 * before legit output with numbers (like the time in your status bar.)
 * Se we have to actually slurp up only those digits that comprise a legal
 * ^C code.
 */
const u_char *read_color_seq (const u_char *start, void *d)
{
	/* 
	 * The proper "attribute" color mapping is for each ^C lvalue.
	 * If the value is -1, then that is an illegal ^C lvalue.
	 */
	static	int	fore_conv[] = {
		 7,  0,  4,  2,  1,  3,  5,  1,		/*  0-7  */
		 3,  2,  6,  6,  4,  5,  0,  7,		/*  8-15 */
		 7, -1, -1, -1, -1, -1, -1, -1, 	/* 16-23 */
		-1, -1, -1, -1, -1, -1,  0,  1, 	/* 24-31 */
		 2,  3,  4,  5,  6,  7, -1, -1,		/* 32-39 */
		-1, -1, -1, -1, -1, -1, -1, -1,		/* 40-47 */
		-1, -1,  0,  1,  2,  3,  4,  5, 	/* 48-55 */
		 6,  7,	-1, -1, -1			/* 56-60 */
	};
	/* 
	 * The proper "attribute" color mapping is for each ^C rvalue.
	 * If the value is -1, then that is an illegal ^C rvalue.
	 */
	static	int	back_conv[] = {
		 7,  0,  4,  2,  1,  3,  5,  1,
		 3,  2,  6,  6,  4,  5,  0,  0,
		 7, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1,
		 0,  1,  2,  3,  4,  5,  6,  7, 
		-1, -1,  0,  1,  2,  3,  4,  5,
		 6,  7, -1, -1, -1
	};

	/*
	 * Some lval codes represent "bold" colors.  That actually reduces
	 * to ^C<non bold> + ^B, so that if you do ^B later, you get the
	 * <non bold> color.  This table indicates whether a ^C code 
	 * turns bold ON or OFF.  (Every color does one or the other)
	 */
	static	int	fore_bold_conv[] =  {
		1,  0,  0,  0,  0,  0,  0,  1,
		1,  1,  0,  1,  1,  1,  1,  0,
		1,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  1,  1,  1,  1,  1,  1,
		1,  1,  0,  0,  0
	};
	/*
	 * Some rval codes represent "blink" colors.  That actually reduces
	 * to ^C<non blink> + ^F, so that if you do ^F later, you get the
	 * <non blink> color.  This table indicates whether a ^C code 
	 * turns blink ON or OFF.  (Every color does one or the other)
	 */
	static	int	back_blink_conv[] = {
		0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  1,  1,  1,  1,  1,  1,
		1,  1,  0,  0,  0
	};

	/* Local variables, of course */
	const 	u_char *	ptr = start;
		int		c1, c2;
		Attribute *	a;
		Attribute	ad;
		int		fg;
		int		val;
		int		noval;

	/* Copy the inward attributes, if provided */
	a = (d) ? (Attribute *)d : &ad;

	/*
	 * If we're passed a non ^C code, dont do anything.
	 */
	if (*ptr != '\003')
		return ptr;

	/*
	 * This is a one-or-two-time-through loop.  We find the maximum 
	 * span that can compose a legit ^C sequence, then if the first 
	 * nonvalid character is a comma, we grab the rhs of the code.  
	 */
	for (fg = 1; ; fg = 0)
	{
		/*
		 * If its just a lonely old ^C, then its probably a terminator.
		 * Just skip over it and go on.
		 */
		ptr++;
		if (*ptr == 0)
		{
			if (fg)
				a->color_fg = a->fg_color = 0;
			a->color_bg = a->bg_color = 0;
			a->bold = a->blink = 0;
			return ptr;
		}

		/*
		 * Check for the very special case of a definite terminator.
		 * If the argument to ^C is -1, then we absolutely know that
		 * this ends the code without starting a new one
		 */
		else if (ptr[0] == '-' && ptr[1] == '1')
		{
			if (fg)
				a->color_fg = a->fg_color = 0;
			a->color_bg = a->bg_color = 0;
			a->bold = a->blink = 0;
			return ptr + 2;
		}

		/*
		 * Further checks against a lonely old naked ^C.
		 */
		else if (!isdigit(ptr[0]) && ptr[0] != ',')
		{
			if (fg)
				a->color_fg = a->fg_color = 0;
			a->color_bg = a->bg_color = 0;
			a->bold = a->blink = 0;
			return ptr;
		}


		/*
		 * Code certainly cant have more than two chars in it
		 */
		c1 = ptr[0];
		c2 = ptr[1];
		val = 0;
		noval = 0;

#define mkdigit(x) ((x) - '0')

		/* Our action depends on the char immediately after the ^C. */
		switch (c1)
		{
			/* These might take one or two characters */
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			{
			    if (c2 >= '0' && c2 <= '9')
			    {
				int	val1;
				int	val2;

			        ptr++;
				val1 = mkdigit(c1);
				val2 = mkdigit(c1) * 10 + mkdigit(c2);

				if (fg)
				{
					if (fore_conv[val2] == -1)
						val = val1;
					else
						val = val2, ptr++;
				}
				else
				{
					if (back_conv[val2] == -1)
						val = val1;
					else
						val = val2, ptr++;
				}
				break;
			    }

			    /* FALLTHROUGH */
			}

			/* These can only take one character */
			case '6':
			case '7':
			case '8':
			case '9':
			{
				ptr++;

				val = mkdigit(c1);
				break;
			}

			/*
			 * Y -> <stop> Y for any other nonnumeric Y
			 */
			default:
			{
				noval = 1;
				break;
			}
		}

		if (noval == 0)
		{
			if (fg)
			{
				a->color_fg = 1;
				a->bold = fore_bold_conv[val];
				a->fg_color = fore_conv[val];
			}
			else
			{
				a->color_bg = 1;
				a->blink = back_blink_conv[val];
				a->bg_color = back_conv[val];
			}
		}

		if (*ptr == ',')
			continue;
		break;
	}

	return ptr;
}

/**************************** STRIP ANSI ***********************************/
/*
 * Used as a translation table when we cant display graphics characters
 * or we have been told to do translation.  A no-brainer, with little attempt
 * at being smart.
 * (JKJ: perhaps we should allow a user to /set this?)
 */
static	u_char	gcxlate[256] = {
  '*', '*', '*', '*', '*', '*', '*', '*',
  '#', '*', '#', '*', '*', '*', '*', '*',
  '>', '<', '|', '!', '|', '$', '_', '|',
  '^', 'v', '>', '<', '*', '=', '^', 'v',
  ' ', '!', '"', '#', '$', '%', '&', '\'',
  '(', ')', '*', '+', ',', '_', '.', '/',
  '0', '1', '2', '3', '4', '5', '6', '7',
  '8', '9', ':', ';', '<', '=', '>', '?',
  '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G',
  'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
  'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
  'Z', 'Y', 'X', '[', '\\', ']', '^', '_',
  '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g',
  'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
  'p', 'q', 'r', 's', 't', 'u', 'v', 'w',
  'x', 'y', 'z', '{', '|', '}', '~', '?',
  'C', 'u', 'e', 'a', 'a', 'a', 'a', 'c',
  'e', 'e', 'e', 'i', 'i', 'i', 'A', 'A',
  'e', 'e', 'e', 'o', 'o', 'o', 'u', 'u',
  'y', 'O', 'U', 'C', '#', 'Y', 'P', 'f',
  'a', 'i', 'o', 'u', 'n', 'N', '^', '^',
  '?', '<', '>', '2', '4', '!', '<', '>',
  '#', '#', '#', '|', '|', '|', '|', '+',
  '+', '+', '+', '|', '+', '+', '+', '+',
  '+', '+', '+', '+', '-', '+', '+', '+',
  '+', '+', '+', '+', '+', '=', '+', '+',
  '+', '+', '+', '+', '+', '+', '+', '+',
  '+', '+', '+', '#', '-', '|', '|', '-',
  'a', 'b', 'P', 'p', 'Z', 'o', 'u', 't',
  '#', 'O', '0', 'O', '-', 'o', 'e', 'U',
  '*', '+', '>', '<', '|', '|', '/', '=',
  '*', '*', '*', '*', 'n', '2', '*', '*'
};

/*
 * State 0 is a "normal, printable character"
 * State 1 is an "eight bit character"
 * State 2 is an "escape character" (\033)
 * State 3 is a "color code character" (\003)
 * State 4 is an "attribute change character"
 * State 5 is a "ROM character" (\022)
 * State 6 is a "character that is never printable."
 * State 7 is a "beep"
 * State 8 is a "tab"
 * State 9 is a "non-destructive space"
 */
static	u_char	ansi_state[256] = {
/*	^@	^A	^B	^C	^D	^E	^F	^G */
	6,	6,	4,	3,	6,	4,	4,	7,  /* 000 */
/*	^H	^I	^J	^K	^L	^M	^N	^O */
	6,	8,	4,	6,	0,	6,	6,	4,  /* 010 */
/*	^P	^Q	^R	^S	^T	^U	^V	^W */
	6,	6,	5,	9,	4,	6,	4,	6,  /* 020 */
/*	^X	^Y	^Z	^[	^\	^]	^^	^_ */
	6,	6,	6,	2,	6,	6,	6,	4,  /* 030 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 040 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 050 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 060 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 070 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 100 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 110 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 120 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 130 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 140 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 150 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 160 */
	0,	0,	0,	0,	0,	0,	0,	0,  /* 170 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 200 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 210 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 220 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 230 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 240 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 250 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 260 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 270 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 300 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 310 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 320 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 330 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 340 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 350 */
	1,	1,	1,	1,	1,	1,	1,	1,  /* 360 */
	1,	1,	1,	1,	1,	1,	1,	1   /* 370 */
};

/*
 * This started off as a general ansi parser, and it worked for stuff that
 * was going out to the display, but it couldnt deal properly with ansi codes,
 * and so when i tried to use it for the status bar, it just all fell to 
 * pieces.  After working it over, i came up with this.  What this does is,
 * (believe it or not) walk through and strip out all the ansi codes in the
 * target string.  Any codes that we recognize as being safe (pretty much
 * just ^[[<number-list>m), are converted back into their logical characters
 * (eg, ^B, ^R, ^_, etc), and everything else is completely blown away.
 *
 * If "width" is not -1, then every "width" printable characters, a \n
 * marker is put into the output so you can tell where the line breaks
 * are.  Obviously, this is optional.  It is used by prepared_display 
 * and $leftpc().
 */

/*
 * These macros help keep 8 bit chars from sneaking into the output stream
 * where they might be stripped out.
 */
#define this_char() (eightbit ? *str : (*str) & 0x7f)
#define next_char() (eightbit ? *str++ : (*str++) & 0x7f)
#define put_back() (str--)
#define nlchar '\n'

u_char *	normalize_string (const u_char *str, int logical)
{
	u_char *	output;
	u_char		chr;
	Attribute	a;
	int 		pos;
	int		maxpos;
	int 		args[10];
	int		nargs;
	int		i, n;
	int		ansi = get_int_var(DISPLAY_ANSI_VAR);
	int		gcmode = get_int_var(DISPLAY_PC_CHARACTERS_VAR);
	int		eightbit = term_eight_bit();
	int		beep_max, beep_cnt = 0;
	int		tab_max, tab_cnt = 0;
	int		nds_max, nds_cnt = 0;
	int		pc = 0;
	int		reverse, bold, blink, underline, altchar, color;
	size_t		(*attrout) (u_char *, Attribute *);

	/* Figure out how many beeps/tabs/nds's we can handle */
	if (!(beep_max  = get_int_var(BEEP_MAX_VAR)))
		beep_max = -1;
	if (!get_int_var(TAB_VAR))
		tab_max = -1;
	else if ((tab_max = get_int_var(TAB_MAX_VAR)) < 0)
		tab_max = -1;
	if (!(nds_max	= get_int_var(ND_SPACE_MAX_VAR)))
		nds_max = -1;
	if (normalize_permit_all_attributes)	/* XXXX */
		reverse = bold = blink = underline = altchar = color = 1;
	else
	{
		reverse 	= get_int_var(INVERSE_VIDEO_VAR);
		bold 		= get_int_var(BOLD_VIDEO_VAR);
		blink 		= get_int_var(BLINK_VIDEO_VAR);
		underline 	= get_int_var(UNDERLINE_VIDEO_VAR);
		altchar 	= get_int_var(ALT_CHARSET_VAR);
		color 		= get_int_var(COLOR_VAR);
	}
	if (logical == 0)
		attrout = display_attributes;	/* prep for screen output */
	else if (logical == 1)
		attrout = logic_attributes;	/* non-screen handlers */
	else if (logical == 2)
		attrout = ignore_attributes;	/* $stripansi() function */
	else if (logical == 3)
		attrout = display_attributes;	/* The status line */

	/* Reset all attributes to zero */
	a.bold = a.underline = a.reverse = a.blink = a.altchar = 0;
	a.color_fg = a.color_bg = a.fg_color = a.bg_color = 0;

	/* 
	 * The output string has a few extra chars on the end just 
	 * in case you need to tack something else onto it.
	 */
	maxpos = strlen(str);
	output = (u_char *)new_malloc(maxpos + 64);
	pos = 0;

	while ((chr = next_char()))
	{
	    if (pos > maxpos)
	    {
		maxpos += 64; /* Extend 64 chars at a time */
		RESIZE(output, unsigned char, maxpos + 64);
	    }

	    switch (ansi_state[chr])
	    {
		/*
		 * State 0 is a normal, printable ascii character
		 */
		case 0:
			output[pos++] = chr;
			pc++;
			break;

		/*
		 * State 1 is a high-bit character that may or may not
		 * need to be translated first.
		 * State 6 is an unprintable character that must be made
		 * unprintable (gcmode is forced to be 1)
		 */
		case 1:
		case 6:
		{
			int my_gcmode = gcmode;

			/*
			 * This is a very paranoid check to make sure that
			 * the 8-bit escape code doesnt elude us.
			 */
			if (chr == 27 + 128)
				chr = '[';

			if (ansi_state[chr] == 6)
				my_gcmode = 1;

			if (normalize_never_xlate)
				my_gcmode = 4;

			switch (my_gcmode)
			{
				/*
				 * In gcmode 5, translate all characters
				 */
				case 5:
				{
					output[pos++] = gcxlate[chr];
					break;
				}

				/*
				 * In gcmode 4, accept all characters
				 */
				case 4:
				{
					output[pos++] = chr;
					break;
				}

				/*
				 * In gcmode 3, accept or translate chars
				 */
				case 3:
				{
					if (termfeatures & TERM_CAN_GCHAR)
						output[pos++] = chr;
					else
						output[pos++] = gcxlate[chr];
					break;
				}

				/*
				 * In gcmode 2, accept or highlight xlate
				 */
				case 2:
				{
					if (termfeatures & TERM_CAN_GCHAR)
						output[pos++] = chr;
					else
					{
						/* <REV> char <REV> */
						a.reverse = !a.reverse;
						pos += attrout(output + pos, &a);
						output[pos++] = gcxlate[chr];
						a.reverse = !a.reverse;
						pos += attrout(output + pos, &a);
					}
					break;
				}

				/*
				 * gcmode 1 is "accept or reverse mangle"
				 * If youre doing 8-bit, it accepts eight
				 * bit characters.  If youre not doing 8 bit
				 * then it converts the char into something
				 * printable and then reverses it.
				 */
				case 1:
				{
					if (termfeatures & TERM_CAN_GCHAR)
						output[pos++] = chr;
					else if ((chr & 0x80) && eightbit)
						output[pos++] = chr;
					else
					{
						a.reverse = !a.reverse;
						pos += attrout(output + pos, &a);
						output[pos++] = 
							(chr | 0x40) & 0x7f;
						a.reverse = !a.reverse;
						pos += attrout(output + pos, &a);
					}
					break;
				}

				/*
				 * gcmode 0 is "always strip out"
				 */
				case 0:
					break;
			}
			pc++;
			break;
		}


		/*
		 * State 2 is the escape character
		 */
		case 2:
		{
		    /*
		     * The next thing we do is dependant on what the character
		     * is after the escape.  Be very conservative in what we
		     * allow.  In general, escape sequences shouldnt be very
		     * complex at this point.
		     * If we see an escape at the end of a string, just mangle
		     * it and dont bother with the rest of the expensive
		     * parsing.
		     */
		    if (!ansi || this_char() == 0)
		    {
			a.reverse = !a.reverse;
			pos += attrout(output + pos, &a);
			output[pos++] = '[';
			a.reverse = !a.reverse;
			pos += attrout(output + pos, &a);

			pc++;
			continue;
		    }

		    switch ((chr = next_char()))
		    {
			/*
			 * All these codes we just skip over.  We're not
			 * interested in them.
			 */

			/*
			 * These are two-character commands.  The second
			 * char is the argument.
			 */
			case ('#') : case ('(') : case (')') :
			case ('*') : case ('+') : case ('$') :
			case ('@') :
			{
				next_char();	/* Skip the argument */
				break;
			}

			/*
			 * These are just single-character commands.
			 */
			case ('7') : case ('8') : case ('=') :
			case ('>') : case ('D') : case ('E') :
			case ('F') : case ('H') : case ('M') :
			case ('N') : case ('O') : case ('Z') :
			case ('l') : case ('m') : case ('n') :
			case ('o') : case ('|') : case ('}') :
			case ('~') : case ('c') :
			{
				break;		/* Dont do anything */
			}

			/*
			 * Swallow up graphics sequences...
			 */
			case ('G'):
			{
				while ((chr = next_char()) != 0 &&
					chr != ':')
					;
				if (chr == 0)
					put_back();
				break;
			}

			/*
			 * Not sure what this is, its not supported by
			 * rxvt, but its supposed to end with an ESCape.
			 */
			case ('P') :
			{
				while ((chr = next_char()) != 0 &&
					chr != 033)
					;
				if (chr == 0)
					put_back();
				break;
			}

			/*
			 * Anything else, we just munch the escape and 
			 * leave it at that.
			 */
			default:
				put_back();
				break;


			/*
			 * Strip out Xterm sequences
			 */
			case (']') :
			{
				while ((chr = next_char()) != 0 && chr != 7)
					;
				if (chr == 0)
					put_back();
				break;
			}

			/*
			 * Now these we're interested in....
			 * (CSI sequences)
			 */
			case ('[') :
			{
	/* <<<<<<<<<<<< */
	/*
	 * Set up the arguments list
	 */
	nargs = 0;
	args[0] = args[1] = args[2] = args[3] = 0;
	args[4] = args[5] = args[6] = args[7] = 0;
	args[8] = args[9] = 0;

	/*
	 * This stuff was taken/modified/inspired by rxvt.  We do it this 
	 * way in order to trap an esc sequence that is embedded in another
	 * (blah).  We're being really really really paranoid doing this, 
	 * but it is for the best.
	 */

	/*
	 * Check to see if the stuff after the command is a "private" 
	 * modifier.  If it is, then we definitely arent interested.
	 *   '<' , '=' , '>' , '?'
	 */
	chr = this_char();
	if (chr >= '<' && chr <= '?')
		next_char();	/* skip it */


	/*
	 * Now pull the arguments off one at a time.  Keep pulling them 
	 * off until we find a character that is not a number or a semicolon.
	 * Skip everything else.
	 */
	for (nargs = 0; nargs < 10; str++)
	{
		n = 0;
		for (n = 0; isdigit(this_char()); next_char())
			n = n * 10 + (this_char() - '0');

		args[nargs++] = n;

		/*
		 * If we run out of code here, then we're totaly confused.
		 * just back out with whatever we have...
		 */
		if (!this_char())
		{
			output[pos] = output[pos + 1] = 0;
			return output;
		}

		if (this_char() != ';')
			break;
	}

	/*
	 * If we find a new ansi char, start all over from the top 
	 * and strip it out too
	 */
	if (this_char() == 033)
		continue;

	/*
	 * Support "spaces" (cursor right) code
	 */
	if (this_char() == 'a' || this_char() == 'C')
	{
		next_char();
		if (nargs >= 1)
		{
		       /*
			* Keep this within reality.
			*/
			if (args[0] > 256)
				args[0] = 256;

			/* This is just sanity */
			if (pos + args[0] > maxpos)
			{
				maxpos += args[0]; 
				RESIZE(output, u_char, maxpos + 64);
			}
			while (args[0]-- > 0)
			{
				if (nds_max > 0 && nds_cnt > nds_max)
					break;

				output[pos++] = ND_SPACE;
				pc++;
				nds_cnt++;
			}
		}
		break;
	}


	/*
	 * The 'm' command is the only one that we honor.  
	 * All others are dumped.
	 */
	if (next_char() != 'm')
		break;


	/*
	 * Walk all of the numeric arguments, plonking the appropriate 
	 * attribute changes as needed.
	 */
	for (i = 0; i < nargs; i++)
	{
	    switch (args[i])
	    {
		case 0:		/* Reset to default */
		{
			a.reverse = a.bold = 0;
			a.blink = a.underline = 0;
			a.altchar = 0;
			a.color_fg = a.color_bg = 0;
			a.fg_color = a.bg_color = 0;
			pos += attrout(output + pos, &a);
			break;
		}
		case 1:		/* bold on */
		{
			if (bold)
			{
				a.bold = 1;
				pos += attrout(output + pos, &a);
			}
			break;
		}
		case 2:		/* dim on -- not supported */
			break;
		case 4:		/* Underline on */
		{
			if (underline)
			{
				a.underline = 1;
				pos += attrout(output + pos, &a);
			}
			break;
		}
		case 5:		/* Blink on */
		case 26:	/* Blink on */
		{
			if (blink)
			{
				a.blink = 1;
				pos += attrout(output + pos, &a);
			}
			break;
		}
		case 6:		/* Blink off */
		case 25:	/* Blink off */
		{
			a.blink = 0;
			pos += attrout(output + pos, &a);
			break;
		}
		case 7:		/* Reverse on */
		{
			if (reverse)
			{
				a.reverse = 1;
				pos += attrout(output + pos, &a);
			}
			break;
		}
		case 21:	/* Bold off */
		case 22:	/* Bold off */
		{
			a.bold = 0;
			pos += attrout(output + pos, &a);
			break;
		}
		case 24:	/* Underline off */
		{
			a.underline = 0;
			pos += attrout(output + pos, &a);
			break;
		}
		case 27:	/* Reverse off */
		{
			a.reverse = 0;
			pos += attrout(output + pos, &a);
			break;
		}
		case 30: case 31: case 32: case 33: case 34: 
		case 35: case 36: case 37:	/* Set foreground color */
		{
			if (color)
			{
				a.color_fg = 1;
				a.fg_color = args[i] - 30;
				pos += attrout(output + pos, &a);
			}
			break;
		}
		case 39:	/* Reset foreground color to default */
		{
			if (color)
			{
				a.color_fg = 0;
				a.fg_color = 0;
				pos += attrout(output + pos, &a);
			}
			break;
		}
		case 40: case 41: case 42: case 43: case 44: 
		case 45: case 46: case 47:	/* Set background color */
		{
			if (color)
			{
				a.color_bg = 1;
				a.bg_color = args[i] - 40;
				pos += attrout(output + pos, &a);
			}
			break;
		}
		case 49:	/* Reset background color to default */
		{
			if (color)
			{
				a.color_bg = 0;
				a.bg_color = 0;
				pos += attrout(output + pos, &a);
			}
			break;
		}

		default:	/* Everything else is not supported */
			break;
	    }
	} /* End of for (handling esc-[...m) */
	/* >>>>>>>>>>> */
			} /* End of escape-[ code handling */
		    } /* End of ESC handling */
		    break;
	        } /* End of case 2 handling */


	        /*
	         * Skip over ^C codes, theyre already normalized
	         * well, thats not totaly true.  We do some mangling
	         * in order to make it work better
	         */
		case 3:
		{
			const u_char 	*end;

			put_back();
			end = read_color_seq(str, (void *)&a);

			/*
			 * XXX - This is a short-term hack to prevent an 
			 * infinite loop.  I need to come back and fix
			 * this the right way in the future.
			 *
			 * The infinite loop can happen when a character
			 * 131 is encountered when eight bit chars is OFF.
			 * We see a character 3 (131 with the 8th bit off)
			 * and so we ask skip_ctl_c_seq where the end of 
			 * that sequence is.  But since it isnt a ^c sequence
			 * it just shrugs its shoulders and returns the
			 * pointer as-is.  So we sit asking it where the end
			 * is and it says "its right here".  So there is a 
			 * need to check the retval of skip_ctl_c_seq to 
			 * actually see if there is a sequence here.  If there
			 * is not, then we just mangle this character.  For
			 * the record, char 131 is a reverse block, so that
			 * seems the most appropriate thing to put here.
			 */
			if (end == str)
			{
				/* Turn on reverse if neccesary */
				a.reverse = !a.reverse;
				pos += attrout(output + pos, &a);
				output[pos++] = ' ';
				a.reverse = !a.reverse;
				pos += attrout(output + pos, &a);

				pc++;
				next_char();	/* Munch it */
				break;
			}

			/* Move to the end of the string. */
			str = end;

			/* Suppress the color if no color is permitted */
			if (!color)
			{
				a.color_fg = a.color_bg = 0;
				a.fg_color = a.bg_color = 0;
				break;
			}

			/* Output the new attributes */
			pos += attrout(output + pos, &a);
			break;
		}

		/*
		 * State 4 is for the special highlight characters
		 */
		case 4:
		{
			put_back();
			switch (this_char())
			{
				case REV_TOG:   
					if (reverse)
						a.reverse = !a.reverse;
					break;
				case BOLD_TOG:  
					if (bold)
						a.bold = !a.bold;
					break;
				case BLINK_TOG:
					if (blink)
						a.blink = !a.blink;
					break;
				case UND_TOG:
					if (underline)
						a.underline = !a.underline;
					break;
				case ALT_TOG:
					if (altchar)
						a.altchar = !a.altchar;
					break;
				case ALL_OFF:
					a.reverse = a.bold = a.blink = 0;
					a.underline = a.altchar = 0;
					a.color_fg = a.color_bg = 0;
					a.bg_color = a.fg_color = 0;
					break;
				default:
					break;
			}

			pos += attrout(output + pos, &a);
			next_char();
			break;
		}

		case 5:
		{
			put_back();
			if (str[0] && str[1] && str[2] && str[3])
			{
				u_char	val = 0;

				next_char();
				val += next_char() - '0';
				val *= 10;
				val += next_char() - '0';
				val *= 10;
				val += next_char() - '0';
				output[pos++] = val;
			}
			else 
			{
				a.reverse = !a.reverse;
				pos += attrout(output + pos, &a);
				output[pos++] = 'R';
				a.reverse = !a.reverse;
				pos += attrout(output + pos, &a);
			}
			pc++;
			break;
		}

		case 7:      /* bell */
		{
			beep_cnt++;
			if ((beep_max == -1) || (beep_cnt > beep_max))
			{
				a.reverse = !a.reverse;
				pos += attrout(output + pos, &a);
				output[pos++] = 'G';
				a.reverse = !a.reverse;
				pos += attrout(output + pos, &a);
				pc++;
			}
			else
				output[pos++] = '\007';

			break;
		}

		case 8:		/* Tab */
		{
			tab_cnt++;
			if (tab_max < 0 || 
			    (tab_max > 0 && tab_cnt > tab_max))
			{
				a.reverse = !a.reverse;
				pos += attrout(output + pos, &a);
				output[pos++] = 'I';
				a.reverse = !a.reverse;
				pos += attrout(output + pos, &a);
				pc++;
			}
			else
			{
				int	len = 8 - (pc % 8);
				for (i = 0; i < len; i++)
				{
					output[pos++] = ' ';
					pc++;
				}
			}
			break;
		}

		case 9:		/* Non-destruct space */
		{
			nds_cnt++;

			/*
			 * Just swallop up any ND's over the max
			 */
			if ((nds_max > 0) && (nds_cnt > nds_max))
				;
			else
				output[pos++] = ND_SPACE;
			break;
		}

		default:
		{
			panic("Unknown normalize_string mode");
			break;
		}
	    } /* End of huge ansi-state switch */
	} /* End of while, iterating over input string */

	/* Terminate the output and return it. */
	if (logical == 0)
	{
		a.bold = a.underline = a.reverse = a.blink = a.altchar = 0;
		a.color_fg = a.color_bg = a.fg_color = a.bg_color = 0;
		pos += attrout(output + pos, &a);
	}
	output[pos] = output[pos + 1] = 0;
	return output;
}

/*
 * prepare_display -- this is a new twist on FireClown's original function.
 * We dont do things quite the way they were explained in the previous 
 * comment that used to be here, so here's the rewrite. ;-)
 *
 * This function is used to break a normalized logical line of display 
 * into some number of physical lines of display.  Again, it is important
 * that 'orig_str' absolutely must have been passed through normalize_string()
 * before calling this function, as it no longer does any interpolation
 * of the results.  (Save for the newline character).
 *
 * The logical line is passed in the 'orig_str' variable, and the width of 
 * the physical display is passed in 'max_cols'.   If 'lused' is not NULL, 
 * then it points at an integer that specifies the maximum number of lines 
 * that should be prepared.  The actual number of lines that are prepared is 
 * stored into 'lused'.  The 'flags' variable specifies some extra options, 
 * the only one of which that is supported right now is "PREPARE_NOWRAP" 
 * which indicates that you want the function to break off the text at 
 * 'max_cols' and not to "wrap" the last partial word to the next line. 
 * ($leftpc() depends on this)
 */
#define SPLIT_EXTENT 40
unsigned char **prepare_display(const unsigned char *orig_str,
                                int max_cols,
                                int *lused,
                                int flags)
{
static 	int 	recursion = 0, 
		output_size = 0;
	int 	pos = 0,            /* Current position in "buffer" */
		col = 0,            /* Current column in display    */
		word_break = 0,     /* Last end of word             */
		indent = 0,         /* Start of second word         */
		firstwb = 0,	    /* Buffer position of second word */
		line = 0,           /* Current pos in "output"      */
		do_indent,          /* Use indent or continued line? */
		newline = 0;        /* Number of newlines           */
static	u_char 	**output = NULL;
const 	u_char	*ptr;
const	u_char	*str;
	u_char 	buffer[BIG_BUFFER_SIZE + 1],
		*cont_ptr,
		*cont = empty_string,
		c,
		*words,
		*pos_copy;
	Attribute	a;

	if (recursion)
		panic("prepare_display() called recursively");
	recursion++;

	/* Get /set's pertaining how to break things up */
	do_indent 	= get_int_var(INDENT_VAR);
	if (!(words 	= get_string_var(WORD_BREAK_VAR)))
		words = ", ";
	if (!(cont_ptr  = get_string_var(CONTINUED_LINE_VAR)))
		cont_ptr = empty_string;

	buffer[0] = 0;

	/* Handle blank or non-existant lines */
	if (!orig_str || !orig_str[0])
		orig_str = " ";

	if (!output_size)
	{
		int 	new_i = SPLIT_EXTENT;
		RESIZE(output, char *, new_i);
		while (output_size < new_i)
			output[output_size++] = 0;
	}

	/*
	 * Start walking through the entire string.
	 */
	str = orig_str;
	for (ptr = str; *ptr && (pos < BIG_BUFFER_SIZE - 8); ptr++)
	{
	    /* Forced newline */
	    if (*ptr == '\n')
	    {
		newline = 1;
		if (indent == 0)
			indent = -1;
		word_break = pos;
	    }

	    /* Is it a space character? */
	    else if (*ptr == ' ' || *ptr == ND_SPACE || strchr(words, *ptr))
	    {
		if (indent == 0)
		{
			indent = -1;
			firstwb = pos;
		}
		word_break = pos;
		if (*ptr != ' ' && ptr[1] && (col + 1 < max_cols))
			word_break++;
		buffer[pos++] = *ptr;
	    }

	    /* Attribute changes -- copy them unmodified. */
	    else if (*ptr == '\006')
	    {
		if (read_attributes(ptr, &a) == 0)
		{
			buffer[pos++] = *ptr++;
			buffer[pos++] = *ptr++;
			buffer[pos++] = *ptr++;
			buffer[pos++] = *ptr++;
			buffer[pos++] = *ptr;
		}
		else
			abort();
		continue;		/* Skip the column check */
	    }

	    else
	    {
		if (indent == -1)
			indent = col;
		buffer[pos++] = *ptr;
	    }


	    /*
	     * Must check for cols >= maxcols+1 becuase we can have a 
	     * character on the extreme screen edge, and we would still 
	     * want to treat this exactly as 1 line, and col has already 
	     * been incremented.
	     */
	    if (col++ >= max_cols || newline)
	    {
		/*
		 * Default the end of line wrap to the last character
		 * we parsed if there were no spaces in the line, or
		 * if we're preparing output that is not to be
		 * wrapped (such as for counting output length.
		 */
		if (word_break == 0 || (flags & PREPARE_NOWRAP))
			word_break = pos;

		/*
		 * XXXX Massive hackwork here.
		 *
		 * Due to some ... interesting design considerations,
		 * if you have /set indent on and your first line has
		 * exactly one word seperation in it, then obviously
		 * there is a really long "word" to the right of the 
		 * first word.  Normally, we would just break the 
		 * line after the first word and then plop the really 
		 * big word down to the second line.  Only problem is 
		 * that the (now) second line needs to be broken right 
		 * there, and we chew up (and lose) a character going 
		 * through the parsing loop before we notice this.  
		 * Not good.  It seems that in this very rare case, 
		 * people would rather not have the really long word 
		 * be sent to the second line, but rather included on 
		 * the first line (it does look better this way), 
		 * and so we can detect this condition here, without 
		 * losing a char but this really is just a hack when 
		 * it all comes down to it.  Good thing its cheap. ;-)
		 */
		if (!*cont && (firstwb == word_break) && do_indent) 
			word_break = pos;

		/*
		 * If we are approaching the number of lines that
		 * we have space for, then resize the master line
		 * buffer so we dont run out.
		 */
		if (line >= output_size - 3)
		{
			int new_i = output_size + SPLIT_EXTENT;
			RESIZE(output, char *, new_i);
			while (output_size < new_i)
				output[output_size++] = 0;
		}

		/*
		 * Now we break off the line at the last space or
		 * last char and copy it off to the master buffer.
		 */
		c = buffer[word_break];
		buffer[word_break] = 0;
		malloc_strcpy((char **)&(output[line++]), buffer);
		buffer[word_break] = c;

		/*
		 * Calculate the continued line marker.  This is
		 * a little bit tricky because we cant figure it out
		 * until after the first line is done.  The first
		 * time through, cont == empty_string, so if !*cont,
		 * we know it has not been initialized.
		 *
		 * So if it has not been initialized and /set indent
		 * is on, and the place to indent is less than a third
		 * of the screen width and /set continued_line is
		 * less than the indented width, then we pad the 
		 * /set continued line value out to the appropriate
		 * width.
		 */
		if (!*cont && do_indent && (indent < (max_cols / 3)) &&
				(strlen(cont_ptr) < indent))
		{
			cont = alloca(indent + 10); /* sb pana */
			sprintf(cont, "%-*s", indent, cont_ptr);
		}
		/*
		 * Otherwise, we just use /set continued_line, 
		 * whatever it is.
		 */
		else if (!*cont && *cont_ptr)
			cont = cont_ptr;

		/*
		 * Skip over all spaces that occur after the break
		 * point, up to the right part of the screen (where
		 * we are currently parsing).  This is what allows
		 * lots and lots of spaces to take up their room.
		 * We let spaces fill in lines as much as neccesary
		 * and if they overflow the line we let them bleed
		 * to the next line.
		 */
		while (buffer[word_break] == ' ' && word_break < pos)
			word_break++;

		/*
		 * At this point, we still have some junk left in
		 * 'buffer' that needs to be moved to the next line.
		 * But of course, all new lines must be prefixed by
		 * the /set continued_line and /set indent stuff, so
		 * we copy off the stuff we have to a temporary
		 * buffer, copy the continued-line stuff into buffer
		 * and then re-append the junk into buffer.  Then we
		 * fix col and pos appropriately and continue parsing 
		 * str...
		 */
		/* 'pos' has already been incremented... */
		buffer[pos] = 0;
		pos_copy = LOCAL_COPY(buffer + word_break);
		display_attributes(buffer, &a);
		strlcpy(buffer, cont, BIG_BUFFER_SIZE / 2);
		strlcat(buffer, pos_copy, BIG_BUFFER_SIZE / 2);

		col = pos = strlen(buffer);
		word_break = 0;
		newline = 0;

		/*
		 * The 'lused' argument allows us to truncate the
		 * parsing at '*lused' lines.  This is most helpful
		 * for the $leftpc() function, which sets a logical
		 * screen width and asks us to "output" one line.
		 */
		if (*lused && line >= *lused)
		{
			*buffer = 0;
			break;
		}
	    } /* End of new line handling */
	} /* End of (ptr = str; *ptr && (pos < BIG_BUFFER_SIZE - 8); ptr++) */

	buffer[pos] = 0;
	if (*buffer)
		malloc_strcpy((char **)&(output[line++]),buffer);

	recursion--;
	new_free(&output[line]);
	*lused = line - 1;
	return output;
}

/*
 * rite: This is the primary display wrapper to the 'output_with_count'
 * function.  This function is called whenever a line of text is to be
 * displayed to an irc window.  It is assumed that the cursor has been
 * placed in the appropriate position before this function is called.
 *
 * This function will "scroll" the target window.  Note that "scrolling"
 * is both a logical and physical act.  The window needs to be told that
 * a line is going to be output, and so it needs to be able to adjust its
 * top_of_display pointer; the hardware terminal also needs to be scrolled
 * so that there is room to put the new text.  scroll_window() handles both
 * of these tasks for us.
 *
 * output_with_count() actually calls putchar_x() for each character in
 * the string, doing the physical output.  It also emits any attribute
 * markers that are in the string.  It does do a clear-to-line, but it does
 * NOT move the cursor away from the end of the line.  We do that after it
 * has returned.
 *
 * This function is used by both irciiwin_display, and irciiwin_repaint.
 * Dont ever 'fold' it in anywhere.
 *
 * The arguments:
 *	window		- The target window for the output
 *	str		- What is to be outputted
 */
static int 	rite (Window *window, const unsigned char *str)
{
	output_screen = window->screen;
	scroll_window(window);

	if (window->visible && foreground && window->display_size)
	{
		output_with_count(str, 1, 1);
		term_cr();
		term_newline();
	}

	window->cursor++;
	return 0;
}

/*
 * This is the main logical/physical output routine.  In its most obvious
 * use, 'cleareol' and 'output' is 1, and it outputs 'str' to the main
 * display (controlled by output_screen), outputting any attribute markers
 * that it finds along the way.  The return value is the number of physical
 * printable characters output.  However, if 'output' is 0, then no actual
 * output is performed, but the counting still takes place.  If 'clreol'
 * is 0, then the rest of the line is not cleared after 'str' has been
 * completely output.  If 'output' is 0, then clreol is ignored.
 */
int 	output_with_count (const unsigned char *str1, int clreol, int output)
{
	int 		beep = 0, 
			out = 0;
	Attribute	a;
	const unsigned char *str;

	for (str = str1; str && *str; str++)
	{
		switch (*str)
		{
			/* Attribute marker */
			case '\006':
			{
				if (read_attributes(str, &a))
					break;
				if (output)
					term_attribute(&a);
				str += 4;
				break;
			}

			/* Terminal beep */
			case '\007':
			{
				beep++;
				break;
			}

			/* Dont ask */
			case '\f':
			{
				if (output)
				{
					a.reverse = !a.reverse;
					term_attribute(&a);
					putchar_x('f');
					a.reverse = !a.reverse;
					term_attribute(&a);
				}
				out++;
				break;
			}

			/* Non-destructive space */
			case ND_SPACE:
			{
				if (output)
					term_right(1);
				out++;		/* Ooops */
				break;
			}

			/* Any other printable character */
			default:
			{
				/*
				 * Note that 'putchar_x()' is safe here
				 * because normalize_string() has already 
				 * removed all of the nasty stuff that could 
				 * end up getting here.  And for those things
				 * that are nasty that get here, its probably
				 * because the user specifically asked for it.
				 */
				if (output)
					putchar_x(*str);
				out++;
				break;
			}
		}
	}

	if (output)
	{
		if (beep)
			term_beep();
		if (clreol)
			term_clear_to_eol();
	}

	return out;
}


/*************************** SCROLLING **********************************/
/*
 * scroll_window: Given a window, this is responsible for making sure that
 * the cursor is placed onto the "next" line.  If the window is full, then
 * it will scroll the window as neccesary.  The cursor is always set to the
 * correct place when this returns.
 */
static void 	scroll_window (Window *window)
{
	if (dumb_mode)
		return;

	/*
	 * If the cursor is beyond the window then we should probably
	 * look into scrolling.
	 */
	if (window->cursor == window->display_size)
	{
		int scroll, i;

		/*
		 * If we ever need to scroll a window that is in scrollback
		 * or in hold_mode, then that means either display_window isnt
		 * doing its job or something else is completely broken.
		 * Probably shouldnt be fatal, but i want to trap these.
		 */
		if (window->holding_something || window->scrollback_point)
			panic("Cant scroll this window!");


		/* Scroll by no less than 1 line */
		if ((scroll = get_int_var(SCROLL_LINES_VAR)) <= 0)
			scroll = 1;

		/* Adjust the top of the logical display */
		for (i = 0; i < scroll; i++)
		{
			window->top_of_display = window->top_of_display->next;
			window->distance_from_display--;
		}

		/* Adjust the top of the physical display */
		if (window->visible && foreground && window->display_size)
		{
			term_scroll(window->top,
				window->top + window->cursor - 1, 
				scroll);
		}

		/* Adjust the cursor */
		window->cursor -= scroll;
	}

	/*
	 * Move to the new line and wipe it
	 */
	if (window->visible && window->display_size)
	{
		window->screen->cursor_window = window;
		term_move_cursor(0, window->top + window->cursor);
		term_clear_to_eol();
	}
}



/* * * * * * * SCREEN UDPATING AND RESIZING * * * * * * * * */
/*
 * repaint_window: redraw the "m"th through the "n"th part of the window.
 * If end_line is -1, then that means clear the display if any of it appears
 * after the end of the scrollback buffer.
 */
void 	repaint_window (Window *w, int start_line, int end_line)
{
	Window *window = (Window *)w;
	Display *curr_line;
	int 	count;
	int 	clean_display = 0;

	if (dumb_mode || !window->visible)
		return;

	if (!window)
		window = current_window;

	if (end_line == -1)
	{
		end_line = window->display_size;
		clean_display = 1;
	}

	curr_line = window->top_of_display;
	for (count = 0; count < start_line; count++)
		curr_line = curr_line->next;

	window->cursor = start_line;
	for (count = start_line; count < end_line; count++)
	{
		rite(window, curr_line->line);

		if (curr_line == window->display_ip)
			break;

		curr_line = curr_line->next;
	}

	/*
	 * If "end_line" is -1, then we're supposed to clear off the rest
	 * of the display, if possible.  Do that here.
	 */
	if (clean_display)
	{
		for (; count < end_line - 1; count++)
		{
			term_clear_to_eol();
			term_newline();
		}
	}

	recalculate_window_cursor(window);	/* XXXX */
}


